from __future__ import absolute_import, print_function
import numpy as np
from tangos.util import timing_monitor
import six
from six.moves import zip
import importlib
import warnings
import functools
from .. import simulation_output_handlers


class HaloPropertiesMetaClass(type):
    # Present to register new subclasses of HaloProperties, so that subclasses can be dynamically
    # instantiated when required based on their cls.names values. Stored as a dictionary so that
    # reloaded classes overwrite their old versions.
    def __init__(cls, name, bases, dict):
        type.__init__(cls, name, bases, dict)
        if hasattr(cls, 'name'):
            warnings.warn("%r defines a name() class method which is deprecated. Use instead a class static variable 'names'."%cls, DeprecationWarning)
            cls.names = cls.name()
        if callable(cls.requires_particle_data):
            warnings.warn("%r defines a requires_particle_data() class method which is deprecated; it should instead be a class static variable", DeprecationWarning)
            cls.requires_particle_data = cls.requires_particle_data()
        if cls.names is not None:
            cls._all_classes.append(cls)

class HaloProperties(six.with_metaclass(HaloPropertiesMetaClass,object)):
    _all_classes = []

    # In child class, defines the most general handler that this property is compatible with.
    # If unchanged, it is compatible with all handlers. Typically it will be appropriate to specify something more
    # restrictive e.g. simulation_output_handlers.pynbody.PynbodyOutputSetHandler
    works_with_handler = simulation_output_handlers.SimulationOutputSetHandler

    # Specifies whether the particle data needs to be provided for this class to perform a calculation; if
    # False, only existing HaloProperties are required by this calculation (see requires_property below).
    requires_particle_data = False

    # Specifies a tuple of names of properties that will be calculated by this class.
    names = None

    @classmethod
    def all_classes(cls):
        return cls._all_classes

    def __init__(self, simulation):
        """Initialise a HaloProperties calculation object

        :param simulation: The simulation from which the properties will be derived
        :type simulation: tangos.core.simulation.Simulation
        """
        self._simulation = simulation
        self.timing_monitor = timing_monitor.TimingMonitor()

    @classmethod
    def index_of_name(cls, name):
        """Returns the index of the named property in the
        results returned from calculate().

        For example, for a BasicHaloProperties object X,
        X.calculate(..)[X.index_of_name("SSC")] returns the SSC.
        """
        name = name.split("(")[0]
        return cls.names.index(name)

    @classmethod
    def no_proxies(self):
        """Returns True if the properties MUST be supplied
        as an actual Halo object rather than the normal
        dictionary-like proxy that is supplied. Only use if
        absolutely necessary; has adverse consequences e.g.
        because uncommitted updates are not reflected into
        your calculate function"""
        return False

    def requires_property(self):
        """Returns a list of existing properties
        required to calculate this property"""
        return []

    def preloop(self, sim, db_timestep):
        """Perform one-per-snapshot calculations, given the loaded simulation data and TimeStep object"""
        pass

    def region_specification(self, db_data):
        """Returns an abstract specification of the region that this halo property is to be calculated on,
        or None if we want the halo particles as defined by the finder.

        See spherical_region.SphericalRegionHaloProperties for an example and useful base class for returning
        everything within the virial radius."""
        return None


    def mark_timer(self, label=None):
        """Called by subfunctions to mark a time"""
        self.timing_monitor.mark(label)


    def accept(self, db_entry):
        for x in self.requires_property():
            if db_entry.get(x, None) is None:
                return False
        return True

    def calculate(self, particle_data, halo_entry):
        """Calculate the properties using the given data

        :param particle_data: The raw particle data, if available
        :type particle_data: pynbody.snapshot.SimSnap (when the pynbody backend is in use, otherwise could be a yt snapshot etc)

        :param halo_entry: The database object associated with the halo, if available
        :type halo_entry: tangos.core.halo.Halo
        :return: All properties as named by names()
        """
        raise NotImplementedError

    def live_calculate(self, halo_entry, *input_values):
        """Calculate the result of a function, using the existing data in the database alone

        :param halo_entry: The database object associated with the halo
        :type halo_entry: tangos.core.halo.Halo

        :param input_values: Input values for the function
        :return: All function values as named by self.names
        """
        if self.requires_particle_data:
            raise(RuntimeError("Cannot live-calculate a property that requires particle data"))
        return self.calculate(None, halo_entry)

    def live_calculate_named(self, name, halo_entry, *input_values):
        """Calculate the result of a function, using the existing data in the database alone

        :param name: The name of the one property to return (which must be one of the values specified by self.names)
        :param halo_entry: The database object associated with the halo
        :type halo_entry: tangos.core.halo.Halo

        :param input_values: Input values for the function
        :return: The single named value
        """
        values = self.live_calculate(halo_entry, *input_values)
        names = self.names
        if isinstance(names, six.string_types):
            return values
        else:
            return values[self.names.index(name)]

    def calculate_from_db(self, db):
        if self.requires_particle_data:
            region_spec =  self.region_specification(self)
            if region_spec:
                halo_particles = db.timestep.load_region(region_spec)
            else:
                halo_particles = db.load()

            preloops_done = getattr(halo_particles.ancestor, "_did_preloop", [])
            halo_particles.ancestor._did_preloop = preloops_done

            if str(self.__class__) not in preloops_done:
                self.preloop(halo_particles.ancestor, db.timestep)
                preloops_done.append(str(self.__class__))
        else:
            halo_particles = None
        return self.calculate(halo_particles, db)

    def plot_x_values(self, for_data):
        """Return a suitable array of x values to match the
        given y values"""
        return np.arange(self.plot_x0(), self.plot_x0() + self.plot_xdelta() * (len(for_data) - 0.5), self.plot_xdelta())

    def plot_x_extent(self):
        return None

    def plot_extent(self):
        return None

    def plot_x0(self):
        return 0

    def plot_xdelta(self):
        return 1.0

    def plot_xlabel(self):
        return None

    def plot_ylabel(self):
        return None

    def plot_yrange(self):
        return None

    def plot_xlog(self):
        return True

    def plot_ylog(self):
        return True

    def plot_clabel(self):
        return None

class TimeChunkedProperty(HaloProperties):
    """TimeChunkedProperty implements a special type of halo property where chunks of a histogram are stored
    at each time step, then appropriately reassembled when the histogram is retrieved."""

    nbins = 1000
    tmax_Gyr = 20.0
    minimum_store_Gyr = 1.0

    @property
    def delta_t(self):
        return self.tmax_Gyr/self.nbins

    @classmethod
    def bin_index(self, time):
        """Convert a time (Gyr) to a bin index in the histogram"""
        index = int(self.nbins*time/self.tmax_Gyr)
        if index<0:
            index = 0
        return index

    @classmethod
    def store_slice(self, time):
        """Tells subclasses which have generated a histogram over all time which slice of that histogram
        they should store."""
        return slice(self.bin_index(time-self.minimum_store_Gyr), self.bin_index(time))

    @classmethod
    def reassemble(cls, property, reassembly_type='major'):
        """Reassemble a histogram by suitable treatment of the merger tree leading up to the current halo.

        This function is normally called by the framework (see Halo.get_data_with_reassembly_options) and you would
        rarely call it directly yourself. From within a live-calculation, it can be accessed using the
        reassemble(halo_property, options...) function. See live_calculation.md for more information.

        :param: property - the halo property for which the reassembly should occur

        :param: reassembly_type - if 'major' (default), return the histogram for the major progenitor branch
                                - if 'sum', return the histogram summed over all progenitors
                                (e.g. this can be used to return SFR histograms that count infalling material as well
                                as the major progenitor)
                                - if 'place', return only the histogram stored at this step but place it within
                                a correctly zero-padded array
                                - if 'raw', return the raw data
        """

        from tangos import relation_finding as rfs

        if reassembly_type=='major':
            return cls._reassemble_using_finding_strategy(property, strategy = rfs.MultiHopMajorProgenitorsStrategy)
        elif reassembly_type=='major_across_simulations':
            return cls._reassemble_using_finding_strategy(property, strategy = rfs.MultiHopMajorProgenitorsStrategy,
                                                          strategy_kwargs = {'target': None})
        elif reassembly_type=='sum':
            return cls._reassemble_using_finding_strategy(property, strategy = rfs.MultiHopAllProgenitorsStrategy)
        elif reassembly_type=='place':
            return cls._place_data(property.halo.timestep.time_gyr, property.data_raw)
        elif reassembly_type=='raw':
            return property.data_raw
        else:
            raise ValueError("Unknown reassembly type")

    @classmethod
    def _place_data(cls, time, raw_data):
        final = np.zeros(cls.bin_index(time))
        end = len(final)
        start = end - len(raw_data)
        final[start:] = raw_data
        return final

    @classmethod
    def _reassemble_using_finding_strategy(cls, property, strategy, strategy_kwargs={}):
        name = property.name.text
        halo = property.halo
        t, stack = halo.calculate_for_descendants("t()", "raw(" + name + ")", strategy=strategy, strategy_kwargs=strategy_kwargs)
        final = np.zeros(cls.bin_index(t[0]))
        previous_time = -1
        for t_i, hist_i in zip(t, stack):
            end = cls.bin_index(t_i)
            start = end - len(hist_i)
            valid = hist_i == hist_i
            if t_i != previous_time:
                # new timestep; overwrite what was there previously
                final[start:end][valid] = hist_i[valid]
            else:
                # same timestep, multiple halos; accumulate
                final[start:end][valid] += hist_i[valid]
            previous_time = t_i
        return final


    def plot_xdelta(cls):
        return cls.tmax_Gyr/cls.nbins

    def plot_xlog(cls):
        return False

    def plot_ylog(cls):
        return False



class LiveHaloProperties(HaloProperties):
    requires_particle_data = False

    def __init__(self, simulation, *args):
        super(LiveHaloProperties, self).__init__(simulation)
        self._nargs = len(args)

    def calculate(self, _, halo):
        return self.live_calculate(halo, *([None]*self._nargs))


class LiveHaloPropertiesInheritingMetaProperties(LiveHaloProperties):
    """LiveHaloProperties which inherit the meta-data (i.e. x0, delta_x values etc) from
    one of the input arguments"""
    def __init__(self, simulation, inherits_from, *args):
        """
        :param simulation: The simulation DB entry for this instance
        :param inherits_from: The HaloProperties description from which the metadata should be inherited
        :type inherits_from: HaloProperties
        """
        super(LiveHaloPropertiesInheritingMetaProperties, self).__init__(simulation)
        self._inherits_from = inherits_from(simulation)

    def plot_x0(self):
        return self._inherits_from.plot_x0()

    def plot_xdelta(self):
        return self._inherits_from.plot_xdelta()

class ProxyHalo(object):

    """Used to return pointers to halos within this snapshot to the database"""

    def __init__(self, value):
        self.value = value

    def __int__(self):
        return int(self.value)



##############################################################################
# UTILITY FUNCTIONS
##############################################################################

def all_property_classes():
    """Return list of all classes derived from HaloProperties"""

    return HaloProperties.all_classes()



def _check_class_provided_name(name):
    if "(" in name or ")" in name:
        raise ValueError("Property names must not include brackets; %s not suitable"%name)

def all_properties(with_particle_data=True):
    """Return list of all properties which can be calculated using classes derived from HaloProperties"""
    classes = all_property_classes()
    pr = []
    for c in classes:
        if c.requires_particle_data and not with_particle_data:
            continue

        name = c.names
        if isinstance(name, six.string_types):
            _check_class_provided_name(name)
            pr.append(name)
        else:
            for name_j in name:
                _check_class_provided_name(name_j)
                pr.append(name_j)

    return pr


def providing_class(property_name, handler_class=None, silent_fail=False):
    """Return property calculator class for given property name when files will be loaded by specified handler.

    If handler_class is None, return "live" properties which can be calculated without particle data"""

    candidates = all_providing_classes(property_name)

    if handler_class is None:
        candidates = list(filter(lambda c: not c.requires_particle_data, candidates))
    else:
        candidates = list(filter(lambda c: issubclass(handler_class, c.works_with_handler), candidates))

    if len(candidates)>=1:
        # return the property which is most specialised
        _sort_by_class_hierarchy(candidates)
        return candidates[0]
    elif silent_fail:
        return None
    else:
        raise NameError("No providing class for property " + property_name)


def all_providing_classes(property_name):
    """Return all the calculator classes for the given property name (possibly multiple, for different handlers)"""
    classes = all_property_classes()
    property_name = property_name.lower()
    candidates = []
    for c in classes:
        name = c.names
        if isinstance(name, tuple) or isinstance(name, list):
            for name_j in name:
                if name_j.lower() == property_name:
                    candidates.append(c)
        elif name.lower() == property_name:
            candidates.append(c)
    return candidates


def _sort_by_class_hierarchy(candidates):
    def cmp(a, b):
        if a is b:
            return 0
        elif issubclass(a, b):
            return 1
        else:
            return -1
    candidates.sort(key=functools.cmp_to_key(cmp))


def providing_classes(property_name_list, handler_class, silent_fail=False):
    """Return classes for given list of property names; see providing_class for details"""
    classes = []
    for property_name in property_name_list:
        cl = providing_class(property_name, handler_class, silent_fail)
        if cl not in classes and cl is not None:
            classes.append(cl)

    return classes

def instantiate_classes(simulation, property_name_list, silent_fail=False):
    """Instantiate appropriate property calculation classes for a given simulation and list of property names"""
    instances = []
    handler_class = type(simulation.get_output_handler())
    for property_identifier in property_name_list:
        instances.append(providing_class(property_identifier, handler_class, silent_fail)(simulation))

    return instances

def instantiate_class(simulation, property_name, silent_fail=False):
    """Instantiate an appropriate property calculation class for a given simulation and property name"""
    handler_class = type(simulation.get_output_handler())
    instance = instantiate_classes(simulation, [property_name],handler_class, silent_fail)
    if len(instance)==0:
        return None
    else:
        return instance[0]

def _import_configured_property_modules():
    from ..config import property_modules
    for pm in property_modules:
        if pm=="": continue
        try:
            importlib.import_module(pm)
        except ImportError:
            warnings.warn("Failed to import requested property module %r. Some properties may be unavailable."%pm,
                          ImportWarning)

_import_configured_property_modules()
from . import intrinsic, live_profiles, pynbody, yt



from __future__ import absolute_import
from .. import core, config
from ..core import Simulation, TimeStep
from ..log import logger
import six
import numpy as np

class SimulationAdderUpdater(object):
    """This class contains the necessary tools to add a new simulation to the database"""

    def __init__(self, simulation_output, session=None):
        """:type simulation_output tangos.simulation_outputs.SimulationOutputSetHandler"""
        self.simulation_output = simulation_output
        if session is None:
            session = core.get_default_session()
        self.session = session
        self.min_halo_particles = config.min_halo_particles

    @property
    def basename(self):
        return self.simulation_output.basename

    def scan_simulation_and_add_all_descendants(self):
        if not self.simulation_exists():
            logger.info("Add new simulation %r", self.basename)
            logger.info("... using the output handler %r", self.simulation_output.handler_class_name())
            self.add_simulation()
        else:
            logger.warn("Simulation already exists %r", self.basename)

        self.add_simulation_properties()

        for ts_filename in self.simulation_output.enumerate_timestep_extensions():
            if not self.timestep_exists_for_extension(ts_filename):
                ts = self.add_timestep(ts_filename)
                self.add_timestep_properties(ts)
                self.add_objects_to_timestep(ts, core.halo.Halo)
                self.add_objects_to_timestep(ts, core.halo.Group)
            else:
                logger.warn("Timestep already exists %r", ts_filename)

    def simulation_exists(self):
        num_matches = self.session.query(Simulation).filter_by(basename=self.basename).count()
        assert num_matches<2, "Consistency problem - more than one simulation with this name exists"
        return num_matches>0

    def timestep_exists_for_extension(self, ts_extension):
        ex = core.get_default_session().query(TimeStep).filter_by(
            simulation=self._get_simulation(),
            extension=ts_extension).first()
        return ex is not None

    def add_timestep(self, ts_extension):
        logger.info("Add timestep %r to simulation %r",ts_extension,self.basename)
        ex = TimeStep(self._get_simulation(), ts_extension)
        return self.session.merge(ex)

    def add_simulation(self):
        sim = Simulation(self.basename)
        self.session.add(sim)
        self.session.commit()

    def add_simulation_properties(self):
        sim = self._get_simulation()
        properties_dict = self.simulation_output.get_properties()
        properties_dict['handler'] = self.simulation_output.handler_class_name()
        for k, v in six.iteritems(properties_dict):
            if k not in list(sim.keys()):
                logger.info("Add simulation property %r",k)
                sim[k] = v
            elif sim[k]!=v:
                logger.info("Update simulation property %r", k)
                sim[k] = v
            else:
                logger.warn("Simulation property %r already exists", k)

        self.session.commit()

    @staticmethod
    def _autoadd_zeros(enumerate_fn):
        # enable enumerate functions to just return halo numbers, adding on the number of dark matter, star and
        # gas particles for backwards compatibility
        def adapted(*args, **kwargs):
            for result in enumerate_fn(*args, **kwargs):
                if not hasattr(result, "__len__"):
                    yield result, 0, 0, 0
                else:
                    yield result
        return adapted

    def add_objects_to_timestep(self, ts, create_class=core.halo.Halo):
        halos = []
        n_tot = []
        enumerator = self._autoadd_zeros(self.simulation_output.enumerate_objects)

        for finder_id, NDM, Nstar, Ngas in enumerator(ts.extension, object_typetag=create_class.tag,
                                                      min_halo_particles=self.min_halo_particles):
            n_tot.append(NDM+Nstar+Ngas)

        database_id = np.zeros(len(n_tot), dtype=int)

        # Sort by total particle number, largest objects first. Use mergesort for sort stability.
        database_id[np.argsort(-np.array(n_tot),kind='mergesort')] = np.arange(len(n_tot)) + 1


        for database_number,(finder_id, NDM, Nstar, Ngas) in zip(database_id,
                                                                 enumerator(ts.extension, object_typetag=create_class.tag,
                                                                            min_halo_particles=self.min_halo_particles)):
            if NDM > self.min_halo_particles or NDM==0:
                h = create_class(ts, database_number, finder_id, NDM, Nstar, Ngas)
                halos.append(h)

        logger.info("Add %d %ss to timestep %r", len(halos), create_class.__name__, ts)
        self.session.add_all(halos)
        self.session.commit()

    def add_timestep_properties(self, ts):
        for key, value in six.iteritems(self.simulation_output.get_timestep_properties(ts.extension)):
            setattr(ts, key, value)



    def _get_simulation(self):
        return self.session.query(Simulation).filter_by(basename=self.basename).first()





from __future__ import absolute_import
from __future__ import print_function
import matplotlib
matplotlib.use('agg')
import pylab as p
from pyramid.view import view_config
from pyramid.compat import escape
import numpy as np
from . import halo_from_request, timestep_from_request, simulation_from_request
from pyramid.response import Response
from six import BytesIO
from ...log import logger
from ... import core
import threading
import time

_matplotlib_lock = threading.RLock()

def decode_property_name(name):
    name = name.replace("_slash_","/")
    return name

def format_array(data, max_array_length=3):
    if len(data)>max_array_length:
        return "Array"
    data_fmt = []
    for d in data:
        data_fmt.append(format_data(data))
    return "["+(", ".join(data_fmt))+"]"

def format_number(data):
    if np.issubdtype(type(data), np.integer):
        return "%d" % data
    elif np.issubdtype(type(data), np.float):
        if abs(data) > 1e5 or abs(data) < 1e-2:
            return "%.2e" % data
        else:
            return "%.2f" % data

def format_data(data, request=None, relative_to=None, max_array_length=3):
    if hasattr(data,'__len__'):
        return format_array(data, max_array_length)
    elif np.issubdtype(type(data), np.number):
        return format_number(data)
    elif isinstance(data, core.Halo):
        return format_halo(data, request, relative_to)
    else:
        return escape(repr(data))



def _relative_description(this_halo, other_halo) :
    if other_halo is None :
        return "null"
    elif this_halo and this_halo.id==other_halo.id:
        return "this"
    elif this_halo and this_halo.timestep_id == other_halo.timestep_id :
        return "%s %d"%(other_halo.tag,other_halo.halo_number)
    elif this_halo and this_halo.timestep.simulation_id == other_halo.timestep.simulation_id :
        return "%s %d at t=%.2e Gyr"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.time_gyr)
    else :
        if (not this_halo) or abs(this_halo.timestep.time_gyr - other_halo.timestep.time_gyr)>0.001:
            return "%s %d in %8s at t=%.2e Gyr"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.simulation.basename,
                                                   other_halo.timestep.time_gyr)
        else:
            return "%s %d in %8s"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.simulation.basename)


def format_halo(halo, request, relative_to=None):
    if relative_to==halo or request is None:
        return _relative_description(relative_to, halo)
    else:
        link = request.route_url('halo_view', simid=halo.timestep.simulation.basename,
                                 timestepid=halo.timestep.extension,
                                 halonumber=halo.basename)
        return "<a href='%s'>%s</a>"%(link, _relative_description(relative_to, halo))

def can_use_in_plot(data):
    return np.issubdtype(type(data), np.number)

def can_use_elements_in_plot(data_array):
    if len(data_array)==0:
        return False
    else:
        return can_use_in_plot(data_array[0])

def can_use_as_filter(data):
    return np.issubdtype(type(data), np.bool) and not np.issubdtype(type(data), np.number) and not hasattr(data,'__len__')

def can_use_elements_as_filter(data_array):
    if len(data_array)==0:
        return False
    else:
        return can_use_as_filter(data_array[0])

@view_config(route_name='calculate_all', renderer='json')
def calculate_all(request):
    ts = timestep_from_request(request)

    try:
        data, db_id = ts.calculate_all(decode_property_name(request.matchdict['nameid']), 'dbid()')
    except Exception as e:
        return {'error': e.message, 'error_class': type(e).__name__}

    return {'timestep': ts.extension, 'data_formatted': [format_data(d, request) for d in data],
           'db_id': [int(x) for x in db_id],  # problems with jsonifying np.int64; needs to be native int?
            'can_use_in_plot': can_use_elements_in_plot(data),
            'can_use_as_filter': can_use_elements_as_filter(data)}

@view_config(route_name='get_property', renderer='json')
def get_property(request):
    halo = halo_from_request(request)

    try:
        result = halo.calculate(decode_property_name(request.matchdict['nameid']))
    except Exception as e:
        return {'error': e.message, 'error_class': type(e).__name__}

    return {'data_formatted': format_data(result, request, halo),
            'can_use_in_plot': can_use_in_plot(result),
            'can_use_as_filter': can_use_as_filter(result)}


def start(request) :
    p.ioff()
    p.clf()
    request.canvas =  p.get_current_fig_manager().canvas

def finish(request, getImage=True) :
    if getImage:
        enter_finish_time = time.time()
        request.canvas.draw()
        draw_time = time.time()
        buffer = BytesIO()
        p.savefig(buffer, format='png')
        end_time = time.time()
        logger.info("Image rendering: matplotlib %.2fs; PNG conversion %.3fs",draw_time-enter_finish_time, end_time-draw_time)

    p.close()

    if getImage :
        return Response(content_type='image/png',body=buffer.getvalue())


def rescale_plot(request):
    logx = request.GET.get('logx',False)
    logy = request.GET.get('logy',False)
    if logx and logy:
        p.loglog()
    elif logx:
        p.semilogx()
    elif logy:
        p.semilogy()

@view_config(route_name='gathered_plot')
def gathered_plot(request):
    ts = timestep_from_request(request)
    name1 = decode_property_name(request.matchdict['nameid1'])
    name2 = decode_property_name(request.matchdict['nameid2'])
    filter = decode_property_name(request.GET.get('filter', ""))
    object_typetag = request.matchdict.get('object_typetag',None)

    if filter!="":
        v1, v2, f = ts.calculate_all(name1, name2, filter)
        v1 = v1[f]
        v2 = v2[f]
    else:
        v1, v2 = ts.calculate_all(name1, name2)
    with _matplotlib_lock:
        start(request)
        p.plot(v1,v2,'k.')
        p.xlabel(name1)
        p.ylabel(name2)
        rescale_plot(request)
        return finish(request)


@view_config(route_name='cascade_plot')
def cascade_plot(request):
    halo = halo_from_request(request)
    name1 = decode_property_name(request.matchdict['nameid1'])
    name2 = decode_property_name(request.matchdict['nameid2'])
    v1, v2 = halo.calculate_for_progenitors(name1, name2)
    with _matplotlib_lock:
        start(request)
        p.plot(v1,v2,'k')
        p.xlabel(name1)
        p.ylabel(name2)
        rescale_plot(request)
        return finish(request)




def image_plot(request, val, property_info):

    log=request.GET.get('logimage',False)
    with _matplotlib_lock:
        start(request)

        if property_info:
            width = property_info.plot_extent()
        else:
            width = 1.0

        if log and len(val.shape)==2:
            data = np.log10(val)
            data[data!=data]=data[data==data].min()

        else:
            data =val

        print(data.min(),data.max(),width)
        if width is not None :
            p.imshow(data,extent=(-width/2,width/2,-width/2,width/2))
        else :
            p.imshow(data)

        if property_info:
            p.xlabel(property_info.plot_xlabel())
            p.ylabel(property_info.plot_ylabel())

        if len(val.shape) is 2 :
            cb = p.colorbar()
            if property_info and property_info.plot_clabel() :
                cb.set_label(property_info.plot_clabel())

        return finish(request)

@view_config(route_name='array_plot')
def array_plot(request):
    halo = halo_from_request(request)
    name = decode_property_name(request.matchdict['nameid'])


    val, property_info = halo.calculate(name, True)

    if len(val.shape)>1:
        return image_plot(request, val, property_info)

    with _matplotlib_lock:
        start(request)

        p.plot(property_info.plot_x_values(val),val)

        if property_info.plot_xlog() and property_info.plot_ylog():
            p.loglog()
        elif property_info.plot_xlog():
            p.semilogx()
        elif property_info.plot_ylog():
            p.semilogy()

        if property_info.plot_xlabel():
            p.xlabel(property_info.plot_xlabel())

        #if property_info.plot_ylabel():
        #    p.ylabel(property_info.plot_ylabel())

        if property_info.plot_yrange():
            p.ylim(*property_info.plot_yrange())

        return finish(request)


from __future__ import absolute_import

import glob
import os
import os.path
import time
import weakref
import re
import numpy as np

pynbody = None # deferred import; occurs when a PynbodyOutputSetHandler is constructed

from . import halo_stat_files, finding
from . import SimulationOutputSetHandler
from .. import config
from ..log import logger
from six.moves import range


_loaded_halocats = {}

class DummyTimeStep(object):
    def __init__(self, filename):
        self.filename = filename


    def __repr__(self):
        return self.filename

    pass


class PynbodyOutputSetHandler(finding.PatternBasedFileDiscovery, SimulationOutputSetHandler):
    def __init__(self, *args, **kwargs):
        super(PynbodyOutputSetHandler, self).__init__(*args, **kwargs)

        import pynbody as pynbody_local

        global pynbody
        pynbody = pynbody_local

        # old versions of pynbody have no __version__!
        pynbody_version = getattr(pynbody, "__version__","0.00")
        assert pynbody_version>="0.42", "Tangos requires pynbody 0.42 or later"

    def _is_able_to_load(self, filepath):
        try:
            f = pynbody.load(filepath)
            if self.quicker:
                logger.warn("Pynbody was able to load %r, but because 'quicker' flag is set we won't check whether it can also load the halo files", filepath)
            else:
                h = f.halos()
            return True
        except (IOError, RuntimeError):
            return False

    def get_timestep_properties(self, ts_extension):
        ts_filename =  self._extension_to_filename(ts_extension)
        f = pynbody.load(ts_filename)
        try:
            time_gyr = f.properties['time'].in_units("Gyr",**f.conversion_context())
        except:
            time_gyr = -1

        results = {'time_gyr': time_gyr, 'redshift': float(f.properties['z']),
                   'available': True}
        return results

    def load_timestep_without_caching(self, ts_extension, mode=None):
        if mode=='partial' or mode is None:
            f = pynbody.load(self._extension_to_filename(ts_extension))
            f.physical_units()
            return f
        elif mode=='server' or mode=='server-partial':
            from ..parallel_tasks import pynbody_server as ps
            return ps.RemoteSnapshotConnection(self._extension_to_filename(ts_extension))
        else:
            raise NotImplementedError("Load mode %r is not implemented"%mode)

    def load_region(self, ts_extension, region_specification, mode=None):
        if mode is None:
            timestep = self.load_timestep(ts_extension)
            return timestep[region_specification]
        elif mode=='server':
            timestep = self.load_timestep(ts_extension)
            return timestep.get_view(region_specification)
        elif mode=='server-partial':
            timestep = self.load_timestep(ts_extension)
            view = timestep.get_view(region_specification)
            load_index = view['remote-index-list']
            logger.info("Partial load %r, taking %d particles",ts_extension,len(load_index))
            f = pynbody.load(self._extension_to_filename(ts_extension), take=load_index)
            f.physical_units()
            return f
        elif mode=='partial':
            raise NotImplementedError("For partial loading to work with custom regions, you need load-mode=server-partial (instead of load-mode=partial)")
        else:
            raise NotImplementedError("Load mode %r is not implemented"%mode)

    def load_object(self, ts_extension, halo_number, object_typetag='halo', mode=None):
        if mode=='partial':
            h = self._construct_halo_cat(ts_extension, object_typetag)
            h_file = h.load_copy(halo_number)
            h_file.physical_units()
            return h_file
        elif mode=='server':
            timestep = self.load_timestep(ts_extension)
            return timestep.get_view(halo_number)
        elif mode=='server-partial':
            timestep = self.load_timestep(ts_extension)
            view = timestep.get_view(halo_number)
            load_index = view['remote-index-list']
            logger.info("Partial load %r, taking %d particles", ts_extension, len(load_index))
            f = pynbody.load(self._extension_to_filename(ts_extension), take=load_index)
            f.physical_units()
            return f
        elif mode is None:
            h = self._construct_halo_cat(ts_extension, object_typetag)
            return h[halo_number]
        else:
            raise NotImplementedError("Load mode %r is not implemented"%mode)

    def load_tracked_region(self, ts_extension, track_data, mode=None):
        f = self.load_timestep(ts_extension)
        indices = self._get_indices_for_snapshot(f, track_data)
        if mode=='partial':
            return pynbody.load(f.filename, take=indices)
        elif mode is None:
            return f[indices]
        else:
            raise NotImplementedError("Load mode %r is not implemented"%mode)


    def _get_indices_for_snapshot(self, f, track_data):
        pt = track_data.particles
        if track_data.use_iord is True:

            dm_part = f.dm[np.in1d(f.dm['iord'], pt)]

            try:
                star_part = f.star[np.in1d(f.star['iord'], pt)]
            except KeyError:
                star_part = f[0:0]

            try:
                gas_part = f.gas[np.in1d(f.gas['iord'], pt)]
            except KeyError:
                gas_part = f[0:0]

            # fx = dm_part.union(star_part)
            # fx = fx.union(gas_part)
            # return fx
            ilist = np.hstack((dm_part.get_index_list(f),
                               star_part.get_index_list(f),
                               gas_part.get_index_list(f)))
            ilist = np.sort(ilist)
            return ilist
        else:
            return pt



    def _construct_halo_cat(self, ts_extension, object_typetag):
        if object_typetag!= 'halo':
            raise ValueError("Unknown object type %r" % object_typetag)
        f = self.load_timestep(ts_extension)
        # amiga grp halo
        h = _loaded_halocats.get(id(f), lambda: None)()
        if h is None:
            h = f.halos()
            if isinstance(h, pynbody.halo.SubfindCatalogue):
                h = f.halos(subs=True)
            _loaded_halocats[id(f)] = weakref.ref(h)
            f._db_current_halocat = h # keep alive for lifetime of simulation
        return h  # pynbody.halo.AmigaGrpCatalogue(f)




    def match_objects(self, ts1, ts2, halo_min, halo_max,
                    dm_only=False, threshold=0.005, object_typetag='halo'):
        if dm_only:
            only_family='dm'
        else:
            only_family=None

        f1 = self.load_timestep(ts1)
        f2 = self.load_timestep(ts2)

        h1 = self._construct_halo_cat(ts1, object_typetag)
        h2 = self._construct_halo_cat(ts2, object_typetag)

        return f1.bridge(f2).fuzzy_match_catalog(halo_min, halo_max, threshold=threshold,
                                                 only_family=only_family, groups_1=h1, groups_2=h2)

    def enumerate_objects(self, ts_extension, object_typetag="halo", min_halo_particles=config.min_halo_particles):
        if self._can_enumerate_objects_from_statfile(ts_extension, object_typetag):
            for X in self._enumerate_objects_from_statfile(ts_extension, object_typetag):
                yield X
        else:
            logger.warn("No halo statistics file found for timestep %r",ts_extension)

            try:
                h = self._construct_halo_cat(ts_extension, object_typetag)
            except:
                logger.warn("Unable to read %ss using pynbody; assuming step has none", object_typetag)
                raise StopIteration

            logger.warn(" => enumerating %ss directly using pynbody", object_typetag)

            istart = 1

            if isinstance(h, pynbody.halo.SubfindCatalogue):
                istart = 0 # subfind indexes from zero

            if hasattr(h, 'precalculate'):
                h.precalculate()


            for i in range(istart, len(h)+istart):
                try:
                    hi = h[i]
                    if len(hi.dm) > min_halo_particles:
                        yield i, len(hi.dm), len(hi.star), len(hi.gas)
                except (ValueError, KeyError) as e:
                    pass

    def get_properties(self):
        timesteps = list(self.enumerate_timestep_extensions())
        if len(timesteps)>0:
            f = self.load_timestep_without_caching(timesteps[0])
            if self.quicker:
                res = self._estimate_resolution_quicker(f)
            else:
                res = self._estimate_resolution(f)

            return {'approx_resolution_kpc': res}

        else:
            return {}

    def _estimate_resolution(self, f):
        f.physical_units()
        if "eps" in f.dm.loadable_keys():
            # Interpret the eps array as a softening, and assume that it is not a comoving softening (as the
            # pynbody units system might naively tell us) but actually already a physical softening. Note that
            # whether or not this is a correct interpretation depends on the code in use, and the flags passed
            # to that code.
            return float(f.dm['eps'].in_units('kpc a').min())
        else:
            # There is no softening information available, so take a best guess as to what a reasonable
            # softening might be as 1/100th of the mean interparticle distance (in the deepest zoom level)
            tot_box_mass = f.dm['mass'].sum()
            min_mass = f.dm['mass'].min()
            frac_mass = min_mass/tot_box_mass
            frac_length = frac_mass ** (1. / 3)
            estimated_eps = 0.01 * frac_length * f.properties['boxsize'].in_units('kpc a', **f.conversion_context())
            return float(estimated_eps)

    def _estimate_resolution_quicker(self, f):
        interparticle_distance = float(f.properties['boxsize'].in_units("kpc a",**f.conversion_context()))/(float(len(f))**(1./3))
        res = 0.01*interparticle_distance
        logger.warn("Because 'quicker' flag is set, estimating res %.2g kpc from the file size; this could be inaccurate",
                    res)
        logger.warn(" -- it will certainly be wrong for e.g. zoom simulations")
        return res


class RamsesHOPOutputSetHandler(PynbodyOutputSetHandler):
    patterns = ["output_0????"]



class GadgetSubfindOutputSetHandler(PynbodyOutputSetHandler):
    patterns = ["snapshot_???"]

    def load_object(self, ts_extension, halo_number, object_typetag='halo', mode=None):
        if mode=='subfind_properties':
            h = self._construct_halo_cat(ts_extension, object_typetag)
            return h.get_halo_properties(halo_number,with_unit=False)
        else:
            return super(GadgetSubfindOutputSetHandler, self).load_object(ts_extension, halo_number, object_typetag, mode)



    def enumerate_timestep_extensions(self):
        base = os.path.join(config.base, self.basename)
        extensions = finding.find(basename=base + "/", patterns=["snapshot_???"])
        for e in extensions:
            if self._is_able_to_load(e):
                yield e[len(base)+1:]


    def _construct_group_cat(self, ts_extension):
        f = self.load_timestep(ts_extension)
        h = _loaded_halocats.get(id(f)+1, lambda: None)()
        if h is None:
            h = f.halos()
            assert isinstance(h, pynbody.halo.SubfindCatalogue)
            _loaded_halocats[id(f)+1] = weakref.ref(h)
            f._db_current_groupcat = h  # keep alive for lifetime of simulation
        return h

    def _construct_halo_cat(self, ts_extension, object_typetag):
        if object_typetag== 'halo':
            return super(GadgetSubfindOutputSetHandler, self)._construct_halo_cat(ts_extension, object_typetag)
        elif object_typetag== 'group':
            return self._construct_group_cat(ts_extension)
        else:
            raise ValueError("Unknown halo type %r" % object_typetag)




class ChangaOutputSetHandler(PynbodyOutputSetHandler):
    flags_include = ["dPhysDenMin", "dCStar", "dTempMax",
                     "dESN", "bLowTCool", "bSelfShield", "dExtraCoolShutoff"]

    patterns = ["*.00???","*.00????"]


    def get_properties(self):
        parent_prop_dict = super(ChangaOutputSetHandler, self).get_properties()

        pfile = self._get_paramfile_path()

        if pfile is None:
            logger.warn("Param file cannot be found - no simulation properties will be available")
            return {}
        else:
            logger.info("Param file is %s", pfile)

        pfile_dict = self._param_file_to_dict(pfile)
        log_path, prop_dict = self._get_log_path(pfile, pfile_dict)

        if log_path:
            prop_dict.update(self._get_properties_from_log(log_path))

        prop_dict.update(self._filter_paramfile_properties(pfile_dict))
        prop_dict.update(parent_prop_dict)

        return prop_dict

    def _get_paramfile_path(self):
        try:
            pfile = self._get_param_file_for_output(self._extension_to_filename(""))
        except RuntimeError:
            pfile = None
        return pfile

    def _get_log_path(self, paramfile_name, paramfile_dict):
        prop_dict = {}
        log_fn = paramfile_dict.get("achOutName","") + ".log"
        log_path = paramfile_name.split("/")[:-1]
        log_path.append(log_fn)
        log_path = "/".join(log_path)
        if os.path.exists(log_path):
            logger.info("Log file is %s", log_path)
        else:
            logger.warn("Cannot find log file (%s)", log_path)
            log_path = None
        return log_path, prop_dict

    def _filter_paramfile_properties(self, pfile_dict):
        filtered_pfile_dict = {}
        for f in self.flags_include:
            if f in pfile_dict:
                filtered_pfile_dict[f] = pfile_dict[f]
        return filtered_pfile_dict

    def _get_properties_from_log(self, log_path):
        prop_dict = {}
        with open(log_path, 'r') as f:
            for l in f:
                if "# Code compiled:" in l:
                    prop_dict["compiled"] = time.strptime(
                        l.split(": ")[1].strip(), "%b %d %Y %H:%M:%S")
                if "# Preprocessor macros: " in l:
                    prop_dict["macros"] = l.split(": ")[1].strip()
                    break
        return prop_dict

    @staticmethod
    def _get_param_file_for_output(output_file):
        """Work out the param file corresponding to the
        specified output"""

        q = "/".join(output_file.split("/")[:-1])
        if len(q) != 0:
            path = "/".join(output_file.split("/")[:-1]) + "/"
        else:
            path = ""

        candidates = glob.glob(path + "*.param")

        if len(candidates) == 0:
            candidates = glob.glob(path + "../*.param")

        if len(candidates) == 0:
            raise RuntimeError("No .param file in " + path + \
                                " (or parent) -- please supply or create tipsy.info manually")

        candidates = [x for x in candidates if "direct" not in x and "mpeg_encode" not in x]

        if len(candidates) > 1:
            raise RuntimeError("Can't resolve ambiguity -- too many param files matching " + \
                                path)

        return candidates[0]

    @staticmethod
    def _param_file_to_dict(param_file):
        f = open(param_file)
        out = {}

        for line in f:
            try:
                s = line.split()
                if s[1] == "=" and "#" not in s[0]:
                    key = s[0]
                    v = s[2]

                    if key[0] == "d":
                        v = float(v)
                    elif key[0] == "i" or key[0] == "n" or key[0] == "b":
                        v = int(v)

                    out[key] = v
            except (IndexError, ValueError):
                pass
        return out



# Live properties suitable for calculations on underlying profiles, e.g. density profiles, mass profiles etc

from . import LiveHaloProperties
import numpy as np

class AtPosition(LiveHaloProperties):
    def __init__(self, simulation, position, array):
        super(AtPosition, self).__init__(simulation)
        self._array_info = array

    names = "at"

    def live_calculate(self, halo, pos, ar):
        x0 = self._array_info.plot_x0()
        delta_x = self._array_info.plot_xdelta()

        # linear interpolation
        i0 = int((pos-x0)/delta_x)
        i1 = i0+1

        i0_loc = float(i0)*delta_x+x0
        i1_weight = (pos-i0_loc)/delta_x
        i0_weight = 1.0-i1_weight

        if i1>=len(ar) or i0<0:
            return None
        else:
            return ar[i0]*i0_weight + ar[i1]*i1_weight

class MaxMinProperty(LiveHaloProperties):
    def __init__(self, simulation, array):
        super(MaxMinProperty, self).__init__(simulation)
        self._array_info = array

    names = "max", "min", "posmax", "posmin"

    def live_calculate(self, halo, array):
        max_, min_ = np.max(array), np.min(array)
        amax, amin = np.argmax(array), np.argmin(array)
        index_to_r = lambda index: index*self._array_info.plot_xdelta()+self._array_info.plot_x0()
        return float(max_), float(min_), index_to_r(amax), index_to_r(amin)



class ConsistentCollection(object):
    """Access attributes of an underlying collection of objects, ensuring they are consistent.

    This class wraps a list of objects and, when getting attributes or values, behaves transparently as though it is
    just one of those objects provided that they all actually return the same. If any individual in the collection
    returns a different value, a ValueError is raised."""

    def __init__(self, objects):
        self._objects = set(objects)

    def __getitem__(self, item):
        values = [x[item] for x in self._objects]
        return self._ensure_consistent(values, item)

    def __getattr__(self, item):
        values = [getattr(x, item) for x in self._objects]
        return self._ensure_consistent(values, item)

    def get(self, item, default=None):
        values = [x.get(item, default) for x in self._objects]
        return self._ensure_consistent(values, item)

    def _ensure_consistent(self, values, item_name):
        if any([v!=values[0] for v in values]):
            raise ValueError("Item %r is not consistent between members of the collection"%item_name)
        return values[0]

def consistent_simulation_from_halos(halos):
    return ConsistentCollection([x.timestep.simulation for x in halos])


from __future__ import absolute_import
from pyramid.config import Configurator


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    config = Configurator(settings=settings)
    config.include('pyramid_jinja2')
    config.include('.models')
    config.include('.routes')
    config.scan()

    from . import crumbs
    config.add_request_method(crumbs.breadcrumbs, 'breadcrumbs', reify=True)


    app = config.make_wsgi_app()

    j_env = config.get_jinja2_environment()
    j_env.globals.update(zip=zip,len=len)

    return app



from __future__ import absolute_import
from sqlalchemy import Column, Integer, ForeignKey, Float, LargeBinary, Boolean
from sqlalchemy.orm import relationship, backref, deferred

from .. import data_attribute_mapper
from .. import Base
from .. import creator
from ..dictionary import DictionaryItem
from ..halo import Halo


class HaloProperty(Base):
    __tablename__ = 'haloproperties'

    id = Column(Integer, primary_key=True)
    halo_id = Column(Integer, ForeignKey('halos.id'))
    # n.b. backref defined below
    halo = relationship(Halo, cascade='none', backref=backref('all_properties'))

    data_float = Column(Float)
    data_array = deferred(Column(LargeBinary))
    data_int = Column(Integer)

    name_id = Column(Integer, ForeignKey('dictionary.id'))
    name = relationship(DictionaryItem)

    deprecated = Column(Boolean, default=False, nullable=False)

    creator = relationship(creator.Creator, backref=backref(
        'properties', cascade='all', lazy='dynamic'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))

    def __init__(self, halo, name, data):
        if isinstance(halo, Halo):
            self.halo = halo
        else:
            self.halo_id = halo

        self.name = name
        self.data = data
        self.creator_id = creator.get_creator_id()

    def __repr__(self):
        if self.deprecated:
            x = "<HaloProperty (deprecated) "
        else:
            x = "<HaloProperty "
        if self.data_float is not None:
            return (x + self.name.text + "=%.2e" % self.data) + " of " + self.halo.short() + ">"
        elif self.data_array is not None:
            return x + self.name.text + " (array) of " + self.halo.short() + ">"
        elif self.data_int is not None:
            return x + self.name.text + "=" + str(self.data_int) + " of " + self.halo.short() + ">"
        else:
            return x + ">"  # shouldn't be in this state

    def data_is_array(self):
        """Return True if data is an array (without loading the array)"""
        return (self.data_int is None) and (self.data_float is None)

    @property
    def data_raw(self):
        return data_attribute_mapper.get_data_of_unknown_type(self)

    @property
    def data(self):
        return self.get_data_with_reassembly_options()

    def get_data_with_reassembly_options(self, *options):
        try:
            cls = self.name.providing_class(type(self.halo.handler))
        except NameError:
            cls = None

        if hasattr(cls, 'reassemble'):
            return cls.reassemble(self, *options)
        else:
            return self.data_raw

    def x_values(self):
        if not self.data_is_array():
            raise ValueError("The data is not an array")
        return self.name.providing_class(type(self.halo.handler))(self.halo.timestep.simulation).plot_x_values(self.data)

    @data.setter
    def data(self, data):
        data_attribute_mapper.set_data_of_unknown_type(self, data)





from . import PynbodyHaloProperties

class Contamination(PynbodyHaloProperties):

    def calculate(self, halo, exist):
        n_heavy = (halo.dm['mass'] > self.min_dm_mass).sum()
        return float(n_heavy) / len(halo.dm)

    def preloop(self, sim, db_timestep):
        self.min_dm_mass = sim.dm['mass'].min()

    names = "contamination_fraction"


from __future__ import absolute_import
from sqlalchemy import Column, Integer, ForeignKey, String, Float, Boolean
from sqlalchemy.orm import relationship, backref

from . import Base, DictionaryItem


class ArrayPlotOptions(Base):
    __tablename__ = 'arrayplotoptions'

    id = Column(Integer, primary_key=True)
    dictionary_id = Column(Integer, ForeignKey('dictionary.id'))
    relates_to = relationship(DictionaryItem, backref=backref(
        'array_plot_options', cascade='all,delete-orphan'))

    labelx = Column(String)
    labely = Column(String)

    x0 = Column(Float, default=0)
    dx = Column(Float, default=1)
    dx_is_logarithmic = Column(Boolean, default=False)

    plot_x_logarithmic = Column(Boolean, default=False)
    plot_y_logarithmic = Column(Boolean, default=False)

    use_range = Column(Boolean, default=False)

    def __init__(self, dictionary_item=None):
        self.relates_to = dictionary_item


#!/usr/bin/env python

from __future__ import absolute_import
from __future__ import print_function
import argparse

import tangos as db
import tangos.core.timestep
import pynbody

def resolve_units(value):
    if(pynbody.units.is_unit(value)):
        return float(value)
    else:
        return value

def process_halos_or_groups(halos_or_groups, use_properties, pynbody_prefix):
    session = db.core.get_default_session()
    objects = []

    first = halos_or_groups.first()
    if first:
        ts_keep_alive = first.timestep.load()

    for h in halos_or_groups:

        pynbody_properties = h.load(mode='subfind_properties')

        for k in use_properties:
            if pynbody_prefix + k in pynbody_properties:
                data = resolve_units(pynbody_properties[pynbody_prefix + k])
                k_id = tangos.core.get_or_create_dictionary_item(session, k)
                objects.append(tangos.core.HaloProperty(h, k_id, data))

        if "sub_parent" in pynbody_properties:
            targ = db.core.get_default_session().query(db.core.halo.Group).filter_by(
                timestep_id=h.timestep_id,
                finder_id=int(pynbody_properties['sub_parent'])).first()
            if targ is not None:
                parent_id = tangos.core.get_or_create_dictionary_item(session, "parent")
                child_id = tangos.core.get_or_create_dictionary_item(session, "child")
                objects.append(tangos.core.HaloLink(h, targ, parent_id))
                objects.append(tangos.core.HaloLink(targ, h, child_id))

    session.add_all(objects)
    return len(objects)

def main():

    parser = argparse.ArgumentParser()
    db.supplement_argparser(parser)

    parser.add_argument('--sims','--for', action='store', nargs='*',
                                metavar='simulation_name',
                                help='Specify a simulation (or multiple simulations) to run on')

    parser.add_argument('properties', action='store', nargs='*',
                            help="The names of the halo/group properties to import from Subfind, or leave blank to import all data")

    parser.add_argument('--backwards', action='store_true',
                            help='Process low-z timesteps first')

    args = parser.parse_args()

    db.process_options(args)


    base_sim = db.sim_query_from_name_list(args.sims)

    all_properties = ["CM","HalfMassRad","VMax","VMaxRad","mass","pos","spin","vel","veldisp", # halos, i.e. subfind subs
                      "mass","mcrit_200","mmean_200","mtop_200","rcrit_200","rmean_200","rtop_200", # groups
                      ]

    use_properties = args.properties
    if len(use_properties)==0:
        use_properties=all_properties


    with db.get_default_session().no_autoflush: # for performance
        for x in base_sim:
            if not isinstance(x.get_output_handler(),
                              tangos.simulation_output_handlers.pynbody.GadgetSubfindOutputSetHandler):
                raise ValueError("import_from_subfind requires the handler to be SubfindOutputSetHandler")
            timesteps = db.core.get_default_session().query(tangos.core.timestep.TimeStep).filter_by(
                simulation_id=x.id, available=True).order_by(tangos.core.timestep.TimeStep.redshift.desc()).all()

            if args.backwards:
                timesteps = timesteps[::-1]


            for ts in timesteps:
                tangos.log.logger.info("Processing timestep %r", ts)

                num_added = process_halos_or_groups(ts.halos, use_properties, "sub_")

                tangos.log.logger.info("Added %d halo properties",num_added)

                num_added = process_halos_or_groups(ts.groups, use_properties, "")
                tangos.log.logger.info("Added %d group properties", num_added)

                db.core.get_default_session().commit()



from __future__ import absolute_import
import sqlalchemy, sqlalchemy.exc
from sqlalchemy import Column, Integer, String

from . import Base, get_default_session

_dict_id = {}  # maps dictionary text -> database ID
_dict_obj = {} # maps session, dictionary text -> database object


class DictionaryItem(Base):
    __tablename__ = 'dictionary'

    id = Column(Integer, primary_key=True)
    text = Column(String, unique=True)

    def __repr__(self):
        return "<DictionaryItem " + self.text + ">"

    def __init__(self, text):
        self.text = text

    def providing_class(self, handler):
        from .. import properties
        return properties.providing_class(self.text, handler)

def get_dict_id(text, default=None, session=None, allow_query=True):
    """Get a DictionaryItem id for text (possibly cached). Raises KeyError if
    no dictionary object exists for the specified text"""

    from . import Session

    if session is None:
        session = Session()
        _dict_id = _get_dict_cache_for_session(get_default_session())
        close_session=True
    else:
        _dict_id = _get_dict_cache_for_session(session)
        close_session=False

    try:
        return _dict_id[text]
    except KeyError:

        if allow_query:
            try:
                obj = session.query(DictionaryItem).filter_by(text=text).first()
            except:
                if default is None:
                    raise
                else:
                    return default
            finally:
                if close_session:
                    session.close()
        else:
            obj = None

        if obj is None:
            if default is None:
                raise
            else:
                return default

        _dict_id[text] = obj.id
        return obj.id

def get_or_create_dictionary_item(session, name):
    """This tries to get the DictionaryItem corresponding to name from
    the database.  If it doesn't exist, it creates a pending
    object. Note that this must be called *while the database is
    locked under the specified session* to prevent duplicate items
    being created"""

    if session not in _dict_obj:
        _dict_obj[session] = {}

    # try to get it from the cache
    obj = _dict_obj[session].get(name, None)

    if obj is not None:
        return obj

    # try to get it from the db
    obj = session.query(DictionaryItem).filter_by(text=name).first()


    if obj is None:
        # try to create it
        try:
            obj = DictionaryItem(name)
            obj = session.merge(obj)
            session.commit()
        except sqlalchemy.exc.IntegrityError:
            session.rollback()
            obj = session.query(DictionaryItem).filter_by(text=name).first()
            if obj is None:
                raise # can't get it from the DB, can't create it from the DB... who knows...

    _dict_obj[session][name] = obj
    return obj

def _get_dict_cache_for_session(session):
    session_dict = _dict_id.get(session, None)
    if session_dict is None:
        session_dict = {}
        for dict_item in session.query(DictionaryItem):
            session_dict[dict_item.text] = dict_item.id

        _dict_id[session] = session_dict

    return session_dict

def get_lexicon(session):
    """Get a list of all strings known in the dictionary table"""
    dict_cache = _get_dict_cache_for_session(session)
    return dict_cache.keys()


from __future__ import absolute_import
from .. import BuiltinFunction, FixedNumericInput
import scipy, scipy.stats
import math
import numpy as np

def single_array_smooth(ar_in, npix=5, one_tailed=False) :
    kern = scipy.stats.norm.pdf(np.linspace(-3,3,npix))

    if one_tailed :
        kern[:math.ceil(float(npix)/2)-1]=0
    kern/=kern.sum()
    # repeat edge values
    ar_in_extended = np.concatenate(([ar_in[0]]*npix,ar_in,[ar_in[-1]]*npix))
    return np.convolve(ar_in_extended, kern, 'same')[npix:-npix]

@BuiltinFunction.register
def array_smooth(halos, vals, smooth_npix):
    return [single_array_smooth(vals_i, smooth_npix) if vals_i is not None else None for vals_i in vals]
array_smooth.set_input_options(1, provide_proxy=True, assert_class = FixedNumericInput)

@BuiltinFunction.register
def element(halos, arrays, index):
    return [ar_i[index] if ar_i is not None else None for ar_i in arrays]
element.set_input_options(1, provide_proxy=True, assert_class = FixedNumericInput)


from __future__ import absolute_import
from __future__ import print_function
import weakref

import numpy as np
from sqlalchemy import Column, Integer, LargeBinary, Boolean, ForeignKey
from sqlalchemy.orm import relationship, backref

from . import Base
from .halo import Tracker
from . import creator
from .simulation import Simulation
from .timestep import TimeStep
from .halo_data import HaloLink
from .dictionary import get_or_create_dictionary_item
from ..log import logger


class TrackData(Base):
    __tablename__ = 'trackdata'

    id = Column(Integer, primary_key=True)
    particle_array = Column(LargeBinary)

    use_iord = Column(Boolean, default=True, nullable=False)
    halo_number = Column(Integer, nullable=False, default=0)
    creator = relationship(creator.Creator, backref=backref(
        'trackdata', cascade='delete', lazy='dynamic'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))

    simulation_id = Column(Integer, ForeignKey('simulations.id'))
    simulation = relationship(Simulation, backref=backref('trackers', cascade='all, delete-orphan, merge',
                                                          lazy='dynamic', order_by=halo_number), cascade='save-update')

    def __init__(self, sim, halo_num=None):
        self.simulation = sim
        self.creator_id = creator.get_creator_id()
        if halo_num is None:
            hs = []
            for h in sim.trackers.all():
                hs.append(h.halo_number)
            if len(hs) > 0:
                halo_num = max(hs) + 1
            else:
                halo_num = 1
        self.halo_number = halo_num
        self.use_iord = True
        self.particles = []

    @property
    def particles(self):
        try:
            return np.frombuffer(self.particle_array, dtype=int)
        except ValueError:
            return np.empty(shape=0, dtype=int)

    @particles.setter
    def particles(self, to):
        self.particle_array = np.asarray(to, dtype=int).data



    def select(self, f, use_iord=True):
        self.use_iord = use_iord
        if use_iord:
            pt = f['iord']
        else:
            pt = f.get_index_list(f.ancestor)
        self.particles = pt

    def __repr__(self):
        return "<TrackData %d of %s, len=%d" % (self.halo_number, repr(self.simulation), len(self.particles))

    def create_objects(self, class_=Tracker, first_timestep=None):
        from . import Session
        session = Session.object_session(self)

        timesteps = self.simulation.timesteps
        timesteps_id = [x.id for x in timesteps]

        if first_timestep is not None:
            first_index = timesteps_id.index(first_timestep.id)
            timesteps = timesteps[first_index:]

        object_typecode = class_.object_typecode

        for ts in timesteps:
            existing_query = session.query(class_).filter_by(halo_number = self.halo_number, timestep_id = ts.id)

            if existing_query.count() == 0:
                h = class_(ts, self.halo_number)
                h.tracker_id = self.id
                session.add(h)
                logger.info("Added a %s to %r",class_.__name__,ts)
            else:
                logger.debug("This tracker is already present in %r",ts)

        session.commit()

    def create_links(self, class_=Tracker):
        from . import Session
        session = Session.object_session(self)
        all_ = session.query(class_).join(TimeStep, TimeStep.id==class_.timestep_id).\
                                          filter((class_.halo_number == self.halo_number) &
                                                 (TimeStep.simulation_id == self.simulation_id)).\
                                          order_by(TimeStep.time_gyr)

        connection_name = get_or_create_dictionary_item(session, 'tracker_connection')
        for h1,h2 in zip(all_[1:],all_[:-1]):
            l1 = HaloLink(h1,h2,connection_name)
            l2 = HaloLink(h2,h1,connection_name)
            session.add_all([l1,l2])
        session.commit()


def update_tracker_halos(sim=None):
    from tangos.core import get_default_session
    from tangos import get_simulation
    from tangos.util.terminalcontroller import heading

    if sim is None:
        x = get_default_session().query(TrackData).all()
    else:
        x = get_simulation(sim).trackers.all()

    for y in x:
        heading(repr(y))
        y.create_objects()


"""simulation_output_handlers sub-package

This sub-package handles returning views of the original simulation data on disk for further processing.
At the moment, the views are always pynbody views, but adaptors for other frameworks could be implemented by
overriding the functionality.
"""

from __future__ import absolute_import
import os, os.path
from .. import config
from ..log import logger
import importlib
import weakref

class DummyTimeStep(object):
    def __init__(self, filename):
        self.filename = filename


    def __repr__(self):
        return self.filename

    pass


_loaded_timesteps = {}


class SimulationOutputSetHandler(object):
    """This class handles the output from a simulation as it resides on disk.

    Subclasses provide implementations for different formats and situations.
    """

    def __init__(self, basename):
        self.basename = self.strip_slashes(basename)
        self.quicker = False # a flag to indicate that corners may be cut in the interest of efficiency

    @classmethod
    def best_matching_handler(cls, basename):
        """Find the best subclass to read in the specified folder of simulation timesteps"""
        return cls

    def enumerate_timestep_extensions(self):
        """Yield the extension of each timestep available on disk"""
        raise NotImplementedError

    def get_properties(self):
        """Returns a dictionary of properties of the simulation"""
        return {}

    def get_timestep_properties(self, ts_extension):
        """Returns a dictionary of properties of the timestep"""
        return {}

    def enumerate_objects(self, ts_extension, object_typetag='halo'):
        """Yield halo_number, NDM, NStar, Ngas for halos in the specified timestep"""
        raise NotImplementedError

    def _enumerate_objects_from_statfile(self, ts_extension, object_typetag):
        """Implementation of enumerate_objects when the information is provided by a file readable
        by the halo_stat_files module.

        Call from subclasses when this behaviour is desired"""
        from . import halo_stat_files
        assert object_typetag=='halo'
        ts = DummyTimeStep(self._extension_to_filename(ts_extension))
        ts.redshift = self.get_timestep_properties(ts_extension)['redshift']

        statfile = halo_stat_files.HaloStatFile(ts)
        logger.info("Reading halos for timestep %r using a stat file", ts)
        for X in statfile.iter_rows("n_dm", "n_star", "n_gas"):
            yield X

    def _can_enumerate_objects_from_statfile(self, ts_extension, object_typetag):
        """Returns True if the objects can be enumerated from a stat file"""
        from . import halo_stat_files
        ts = DummyTimeStep(self._extension_to_filename(ts_extension))
        ts.redshift = self.get_timestep_properties(ts_extension)['redshift']
        if object_typetag!='halo':
            return False

        try:
            halo_stat_files.HaloStatFile(ts)
            return True
        except IOError:
            return False

    def load_timestep(self, ts_extension, mode=None):
        """Returns an object that connects to the data for a timestep on disk -- possibly a version cached in
        memory"""
        ts_filename = self._extension_to_filename(ts_extension)
        stored_timestep = _loaded_timesteps.get(ts_filename, lambda: None)()
        if stored_timestep is not None:
            return stored_timestep
        else:
            data = self.load_timestep_without_caching(ts_extension, mode=mode)
            _loaded_timesteps[ts_filename] = weakref.ref(data)
            return data

    def load_region(self, ts_extension, region_specification, mode=None):
        """Returns an object that connects to the data for a timestep on disk, filtered using the
        specified region specification. Acceptable region specifications are output handler dependent.

        The returned object may be a sub-view of a cached in-memory timestep."""
        raise NotImplementedError

    def load_timestep_without_caching(self, ts_extension, mode=None):
        """Creates and returns an object that connects to the data for a timestep on disk"""
        raise NotImplementedError

    def load_object(self, ts_extension, halo_number, object_typetag='halo', mode=None):
        """Creates and returns an object that connects to the data for a halo on disk.

        :arg ts_extension - the timestep path (relative to the simulation basename)

        :arg halo_number - the halo number in the raw halo finder output

        :arg object_typetag - the type of halo catalogue (normally 'halo' or, for Subfind, might be 'group' to get
                                                     the top-level groups)

        :arg mode - sets a method for loading the halo, which is dependent on being implemented by the underlying
                    file manager. Current options (for the default pynbody handler) are

                     * None: load entire snapshot from disk, then return halo particles
                     * 'partial': load only the halo particles from disk
                     * 'server': use a remote server model to retrieve the particles
        """
        raise NotImplementedError

    def load_tracked_region(self, ts_extension, track_data, mode=None):
        """Creates and returns an object that connects to the on-disk data for the specified tracked region.

        :arg mode - sets a method for loading the tracked region; see load_object mode for more information"""
        raise NotImplementedError


    @classmethod
    def handler_class_name(cls):
        module = cls.__module__
        if module.startswith(SimulationOutputSetHandler.__module__):
            submodule = module[len(SimulationOutputSetHandler.__module__)+1:]
            return submodule+"."+cls.__name__
        else:
            return module+"."+cls.__name__

    @staticmethod
    def strip_slashes(name):
        """Strip trailing and leading slashes from relative path"""
        if len(name) == 0: return name
        while name[0] == "/":
            name = name[1:]
            if len(name) == 0: return name
        while name[-1] == "/":
            name = name[:-1]
            if len(name) == 0: return name
        return name

    def _extension_to_filename(self, ts_extension):
        """Given the timestep extension, form the path to the actual file"""
        # Explicit str cast - pynbody seems inconsistent about how it responds to unicode filenames
        return str(os.path.join(config.base, self.basename, ts_extension))


def get_named_handler_class(handler):
    """Get a SimulationOutputSetHandler identified by the given name.

    The name is of the format submodule.ClassName

    :rtype SimulationOutputSetHandler"""
    try:
        output_module = importlib.import_module('.'+handler.split('.')[0],__name__)
    except ImportError:
        output_module = importlib.import_module(handler.split('.')[0])
    output_class = getattr(output_module, handler.split('.')[1])
    return output_class



from pyramid.view import notfound_view_config


@notfound_view_config(renderer='templates/404.jinja2')
def notfound_view(request):
    request.response.status = 404
    return {}

def includeme(config):
    config.add_static_view('static', 'static', cache_max_age=3600)

    config.add_route('autocomplete_words', '/autocomplete_words.json')
    config.add_route('simulation_list', '/')
    config.add_route('simulation_view', '/{simid}')
    config.add_route('timestep_view', '/{simid}/{timestepid}')
    config.add_route('halo_view', '/{simid}/{timestepid}/{halonumber}')
    config.add_route('halo_later', '/{simid}/{timestepid}/{halonumber}/later/{n}')
    config.add_route('halo_earlier', '/{simid}/{timestepid}/{halonumber}/earlier/{n}')
    config.add_route('halo_in', '/{simid}/{timestepid}/{halonumber}/in/{n}')
    config.add_route('calculate_all', '/{simid}/{timestepid}/gather/{nameid}.json')
    config.add_route('get_property', '/{simid}/{timestepid}/{halonumber}/{nameid}.json')
    config.add_route('gathered_plot', '/{simid}/{timestepid}/{nameid1}/vs/{nameid2}.png')
    config.add_route('cascade_plot', '/{simid}/{timestepid}/{halonumber}/{nameid1}/vs/{nameid2}.png')
    config.add_route('array_plot', '/{simid}/{timestepid}/{halonumber}/{nameid}.png')
    config.add_route('merger_tree', '/{simid}/{timestepid}/{halonumber}/merger/tree.json')



from __future__ import absolute_import
import tangos
import pyramid.httpexceptions as exc

def halo_from_request(request):
    ts = timestep_from_request(request)
    try:
        halo = ts[request.matchdict['halonumber']]
    except (KeyError, ValueError):
        raise exc.HTTPNotFound()
    if halo is None:
        raise exc.HTTPNotFound()
    return halo

def timestep_from_request(request):
    sim = simulation_from_request(request)
    try:
        ts = tangos.get_timestep(request.matchdict['timestepid'], request.dbsession, sim)
    except RuntimeError:
        raise exc.HTTPNotFound()
    return ts

def simulation_from_request(request):
    try:
        sim = tangos.get_simulation(request.matchdict['simid'], request.dbsession)
    except RuntimeError:
        raise exc.HTTPNotFound()
    return sim



"""Helper classes for defining translations between .stat file of different formats."""

class Function(object):
    """Define a column which is actually a function of other columns"""
    def __init__(self, fn, *input_arg_names):
        self.fn = fn
        self.input_arg_names = input_arg_names

    def __call__(self, raw_input_names, raw_input_values):
        input_args = [raw_input_values[raw_input_names.index(name)] for name in self.input_arg_names]
        return self.fn(*input_args)

    def inputs(self):
        return self.input_arg_names

class Rename(object):
    """Define a column by renaming an existing column"""
    def __init__(self, name):
        self.name = name

    def __call__(self, raw_input_names, raw_input_values):
        return raw_input_values[raw_input_names.index(self.name)]

    def inputs(self):
        return [self.name]




from __future__ import absolute_import
import argparse

import tangos as db
import tangos.core
from .. import config
from tangos import parallel_tasks
from tangos import core
import sqlalchemy, sqlalchemy.orm
from tangos.log import logger
import numpy as np
from six.moves import zip


class GenericLinker(object):
    def __init__(self, session=None):
        self.session = session or core.get_default_session()

    def parse_command_line(self, argv=None):
        parser = self._create_argparser()
        self.args = parser.parse_args(argv)
        db.process_options(self.args)

    def _create_argparser(self):
        parser = argparse.ArgumentParser()
        db.supplement_argparser(parser)
        parser.add_argument("--verbose", action="store_true",
                            help="Print extra information")
        parser.add_argument("--force", action="store_true",
                            help="Generate links even if they already exist for those timesteps")
        parser.add_argument('--type', action='store', type=str, dest='type_', default='halo',
                            help="Secify the object type to run on by tag name (or integer). "
                                 "Can be halo (default), group, or BH.")
        parser.add_argument("--hmax", action="store", type=int, default=None,
                            help="Specify the maximum number of halos per snapshot")
        parser.add_argument('--backwards', action='store_true',
                            help='Process in reverse order (low-z first)')
        parser.add_argument('--dmonly', action='store_true',
                            help='only match halos based on DM particles. Much more memory efficient, but currently only works for Rockstar halos')
        return parser

    def run_calculation_loop(self):

        pair_list = self._generate_timestep_pairs()

        if len(pair_list)==0:
            logger.error("No timesteps found to link")
            return

        pair_list = parallel_tasks.distributed(pair_list)

        object_type = core.halo.Halo.object_typecode_from_tag(self.args.type_)

        for s_x, s in pair_list:
            logger.info("Linking %r and %r",s_x,s)
            if self.args.force or self.need_crosslink_ts(s_x, s, object_type):
                self.crosslink_ts(s_x, s, 0, self.args.hmax, self.args.dmonly, object_typecode=object_type)

    def _generate_timestep_pairs(self):
        raise NotImplementedError("No implementation found for generating the timestep pairs")

    def get_halo_entry(self, ts, halo_number):
        h = ts.halos.filter_by(finder_id=halo_number).first()
        return h

    def need_crosslink_ts(self, ts1, ts2, object_typecode=0):
        num_sources = ts1.halos.count()
        num_targets = ts2.halos.count()
        if num_targets == 0:
            logger.warn("Will not link: no halos in target timestep %r", ts2)
            return False
        if num_sources == 0:
            logger.warn("Will not link: no halos in source timestep %r", ts1)
            return False

        halo_source = sqlalchemy.orm.aliased(core.halo.Halo, name="halo_source")
        halo_target = sqlalchemy.orm.aliased(core.halo.Halo, name="halo_target")
        same_d_id = core.dictionary.get_or_create_dictionary_item(self.session, "ptcls_in_common").id
        exists = self.session.query(core.halo_data.HaloLink).join(halo_source, core.halo_data.HaloLink.halo_from). \
                    join(halo_target, core.halo_data.HaloLink.halo_to). \
                    filter(halo_source.timestep_id == ts1.id, halo_target.timestep_id == ts2.id,
                           halo_source.object_typecode == object_typecode,
                           halo_target.object_typecode == object_typecode,
                        core.halo_data.HaloLink.relation_id == same_d_id).count() > 0
        self.session.commit()

        if exists:
            logger.warn("Will not link: links already exist between %r and %r", ts1, ts2)
            return False
        return True

    def create_db_objects_from_catalog(self, cat, finder_id_to_halos_1, finder_id_to_halos_2, same_d_id):
        items = []
        missing_db_object = 0
        for i, possibilities in enumerate(cat):
            h1 = finder_id_to_halos_1.get(i, None)
            for cat_i, weight in possibilities:
                h2 = finder_id_to_halos_2.get(cat_i, None)

                if h1 is not None and h2 is not None:
                    items.append(core.halo_data.HaloLink(h1, h2, same_d_id, weight))
                else:
                    missing_db_object += 1

        if missing_db_object > 0:
            logger.warn("%d link(s) could not be identified because the halo objects do not exist in the DB",
                        missing_db_object)
        return items

    def make_finder_id_to_halo_map(self, ts, object_typecode):
        halos = ts.objects.filter_by(object_typecode=object_typecode).all()
        halos_map = {h.finder_id: h for h in halos}
        return halos_map

    def crosslink_ts(self, ts1, ts2, halo_min=0, halo_max=None, dmonly=False, threshold=config.default_linking_threshold, object_typecode=0):
        """Link the halos of two timesteps together

        :type ts1 tangos.core.TimeStep
        :type ts2 tangos.core.TimeStep"""
        logger.info("Gathering halo information for %r and %r", ts1, ts2)
        halos1 = self.make_finder_id_to_halo_map(ts1, object_typecode)
        halos2 = self.make_finder_id_to_halo_map(ts2, object_typecode)

        with parallel_tasks.RLock("create_db_objects_from_catalog"):
            same_d_id = core.dictionary.get_or_create_dictionary_item(self.session, "ptcls_in_common")
            self.session.commit()

        output_handler_1 = ts1.simulation.get_output_handler()
        output_handler_2 = ts2.simulation.get_output_handler()
        if not isinstance(output_handler_1, type(output_handler_2)):
            logger.error("Timesteps %r and %r cannot be crosslinked; they are using incompatible file readers",
                         ts1, ts2)
            return

        # keep the files alive throughout (so they are not garbage-collected after the first match_objects):
        snap1 = ts1.load()
        snap2 = ts2.load()

        try:
            cat = output_handler_1.match_objects(ts1.extension, ts2.extension, halo_min, halo_max, dmonly, threshold,
                                               core.halo.Halo.object_typetag_from_code(object_typecode))
            back_cat = output_handler_2.match_objects(ts2.extension, ts1.extension, halo_min, halo_max, dmonly, threshold,
                                               core.halo.Halo.object_typetag_from_code(object_typecode))
        except Exception as e:
            if isinstance(e, KeyboardInterrupt):
                raise
            logger.exception("Exception during attempt to crosslink timesteps %r and %r", ts1, ts2)
            return

        with self.session.no_autoflush:
            logger.info("Gathering links for %r and %r", ts1, ts2)
            items = self.create_db_objects_from_catalog(cat, halos1, halos2, same_d_id)
            logger.info("Identified %d links between %r and %r", len(items), ts1, ts2)
            items_back = self.create_db_objects_from_catalog(back_cat, halos2, halos1, same_d_id)
            logger.info("Identified %d links between %r and %r", len(items_back), ts2, ts1)

        with parallel_tasks.RLock("create_db_objects_from_catalog"):
            logger.info("Preparing to commit links for %r and %r", ts1, ts2)
            self.session.add_all(items)
            self.session.add_all(items_back)
            self.session.commit()
        logger.info("Finished committing total of %d links for %r and %r", len(items)+len(items_back), ts1, ts2)

class TimeLinker(GenericLinker):
    def _generate_timestep_pairs(self):
        logger.info("generating pairs of timesteps")
        base_sim = db.sim_query_from_name_list(self.args.sims)
        pairs = []
        for x in base_sim:
            ts = self.session.query(tangos.core.timestep.TimeStep).filter_by(
                simulation_id=x.id, available=True).order_by(tangos.core.timestep.TimeStep.redshift.desc()).all()
            for a, b in zip(ts[:-1], ts[1:]):
                pairs.append((a, b))
        if self.args.backwards:
            pairs = pairs[::-1]
        return pairs

    def _create_argparser(self):
        parser = super(TimeLinker, self)._create_argparser()
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')
        return parser

class CrossLinker(GenericLinker):
    def _create_argparser(self):
        parser = super(CrossLinker, self)._create_argparser()
        parser.add_argument('sims', action='store', nargs=2,
                            metavar=('name1', 'name2'),
                            help='The two simulations (or, optionally, two individual timesteps) to crosslink')
        return parser

    def _generate_timestep_pairs(self):
        obj1 = db.get_item(self.args.sims[0])
        obj2 = db.get_item(self.args.sims[1])
        if isinstance(obj1, core.TimeStep) and isinstance(obj2, core.TimeStep):
            return [[obj1, obj2]]
        elif isinstance(obj1, core.Simulation) and isinstance(obj2, core.Simulation):
            return self._generate_timestep_pairs_from_sims(obj1, obj2)
        else:
            raise ValueError("No way to link %r to %r"%(obj1, obj2))

    def _generate_timestep_pairs_from_sims(self, sim1, sim2):
        assert sim1 != sim2, "Can't link simulation to itself"

        logger.info("Match timesteps of %r to %r",sim1,sim2)

        ts1s = sim1.timesteps
        ts2s = sim2.timesteps

        pairs = []
        for ts1 in ts1s:
            ts2 = self._get_best_timestep_matching(ts2s, ts1)
            pairing_is_mutual = (self._get_best_timestep_matching(ts1s,ts2)==ts1)
            if pairing_is_mutual:
                logger.info("Pairing timesteps: %r and %r", ts1, ts2)
                pairs+=[(ts1,ts2)]
            else:
                logger.warn("No pairing found for timestep %r",ts1)

        return pairs

    def _get_best_timestep_matching(self, list_of_candidates, timestep_to_match):
        return min(list_of_candidates, key=lambda ts2: abs(ts2.time_gyr - timestep_to_match.time_gyr))



from __future__ import absolute_import
from .. import core
from ..util import consistent_collection

from .multi_hop import MultiHopStrategy
from .one_hop import HopStrategy
from six.moves import range
from six.moves import zip

import sqlalchemy
from sqlalchemy import func, orm

class MultiSourceMultiHopStrategy(MultiHopStrategy):
    """A variant of MultiHopStrategy that finds halos corresponding to multiple start points.

    Note that the behaviour is necessarily somewhat different to the other classes which start from a single
    halo. Specifically, a target *must* be specified, and the direction of the hops to follow is inferred
    from the nature of the target.

    Additionally, as soon as any halo is "matched" in the target, the entire query is stopped. In other words,
    this class assumes that the number of hops is the same to reach all target halos."""

    def __init__(self, halos_from, target, **kwargs):
        """Construct a strategy for finding Halos via multiple "hops" along HaloLinks from multiple start-points

        :param halos_from: a list of all halos to start from.
        :param target: a TimeStep or Simulation object to target.
        :param one_match_per_input: if True (default), return one halo per starting point in order.
                                  The returned halo in each case should be the one with the
                                  highest weight link (i.e. the major progenitor or similar)

                                  if False, *all* linked halos are returned and the caller has to figure out
                                  which one belongs to which starting halo

        Other parameters are passed onto an underlying MultiHopStrategy. However note that the order_by parameter
        has no effect unless one_match_per_input is False.
        """
        directed = self._infer_direction(halos_from, target)
        kwargs["target"] = target
        kwargs["directed"] = directed

        self._return_only_highest_weights = kwargs.pop('one_match_per_input', True)

        # For 'backwards' or 'forwards' searches (basically major progenitors or descendants), keep only the
        # strongest link at each _step_ rather than waiting to the end to select the highest weight.
        # This makes sure one never "hops" from one branch to another (see
        # test_hop_strategy.test_major_progenitor_from_minor_progenitor for an example that exposes
        # this former bug). The actual implementation of the per-step restriction is in the override to
        # _supplement_halolink_query_with_filter, below.
        self._keep_only_highest_weights_per_hop = (directed == "forwards" or directed == "backwards")
        self._keep_only_highest_weights_per_hop&=self._return_only_highest_weights

        super(MultiSourceMultiHopStrategy, self).__init__(halos_from[0], **kwargs)
        self._all_halo_from = halos_from

    def _infer_direction(self, halos_from, target):
        if isinstance(target, core.simulation.Simulation):
            return "across"
        elif isinstance(target, core.timestep.TimeStep):
            collected_halos = consistent_collection.ConsistentCollection(halos_from)
            if collected_halos.timestep.simulation_id!=target.simulation_id:
                return "across"
            elif collected_halos.timestep.time_gyr<target.time_gyr:
                return "forwards"
            else:
                return "backwards"


    def _seed_temp_table(self):
        insert_dictionaries = []
        for i,halo_from in enumerate(self._all_halo_from):
            insert_dictionaries.append({'halo_from_id': halo_from.id, 'halo_to_id': halo_from.id, 'weight': 1.0,
                                        'nhops': 0, 'source_id': i})

        self._connection.execute(self._table.insert(), insert_dictionaries)

    def _generate_next_level_prelim_links(self, from_nhops=0):
        all = self.session.query(self._link_orm_class).filter(self._link_orm_class.halo_to_id!=self._link_orm_class.halo_from_id).all()
        source_ids = [x.source_id for x in all]
        num_sources = len(self._all_halo_from)
        if self._should_halt():
            return 0
        else:
            return super(MultiSourceMultiHopStrategy, self)._generate_next_level_prelim_links(from_nhops)

    def _supplement_halolink_query_with_filter(self, query, table=None):
        query = super(MultiSourceMultiHopStrategy, self)._supplement_halolink_query_with_filter(query,table)

        if self._keep_only_highest_weights_per_hop:

            query = self._extract_max_weight_rows_from_query(query, table)

        return query

    def _extract_max_weight_rows_from_query(self, query, table):
        if table is None:
            table = core.halo_data.HaloLink.__table__

        # return only the highest weight link from each halo
        # (the following join is basically a work-around for the lack of an 'argmax' type functionality in sql)
        subq = query.add_columns(func.max(table.c.weight).label("max_weight")). \
            group_by(table.c.halo_from_id).subquery()
        query = query.join(subq, sqlalchemy.and_(table.c.halo_from_id == subq.c.halo_from_id,
                                                 table.c.weight == subq.c.max_weight))
        return query

    def _should_halt(self):
        return self.query.count()>0

    def _order_by_clause(self, halo_alias, timestep_alias):
        return [self._link_orm_class.source_id] \
               + super(MultiSourceMultiHopStrategy, self)._order_by_clause(halo_alias, timestep_alias)

    def all(self):
        results = self._get_query_all()
        if self._return_only_highest_weights:
            # query results include a source_id column which we now wish to ignore (see
            # _query_ordered below for more information)
            return [x[1].halo_to if x[1] is not None else None
                    for x in results]
        else:
            return [x.halo_to for x in results]

    @property
    def _query_ordered(self):
        query = super(MultiSourceMultiHopStrategy, self)._query_ordered
        if self._return_only_highest_weights:
            query = self._extract_max_weight_rows_from_query(query, self._table)

            # we now need to restructure the results such that they appear in the same order as the input
            # halo list, and a NULL result is returned for 'missing' halos. This way, the result is
            # guaranteed to be in 1-1 correspondence with the input.
            #
            # We do this by joining to the initial seeds in the temp table
            #
            # the query has to explicitly include the source_id, otherwise the sqlalchemy dedup
            # process removes duplicates rows (e.g. if there are several null results, only the
            # first will be returned!) resulting in a query result that is too short and unrecoverable
            # errors in functions that rely on getting back a 1-1 mapping

            original_query_alias = query.subquery()
            original_orm_alias = orm.aliased(self._link_orm_class, original_query_alias)
            source_ids = orm.aliased(self._link_orm_class)

            sources_query = self.session.query(source_ids.source_id,original_orm_alias).\
                select_from(source_ids).\
                filter(source_ids.nhops==0).order_by(source_ids.source_id)

            query = sources_query.outerjoin(original_orm_alias,
                                            source_ids.source_id==original_orm_alias.source_id)



        return query




from __future__ import absolute_import
from __future__ import print_function
import sys
import re
import math
from six import StringIO
from six.moves import range
from six.moves import zip


class TerminalController:

    """
    A class that can be used to portably generate formatted output to
    a terminal.  

    `TerminalController` defines a set of instance variables whose
    values are initialized to the control sequence necessary to
    perform a given action.  These can be simply included in normal
    output to the terminal:

        >>> term = TerminalController()
        >>> print 'This is '+term.GREEN+'green'+term.NORMAL

    Alternatively, the `render()` method can used, which replaces
    '${action}' with the string required to perform 'action':

        >>> term = TerminalController()
        >>> print term.render('This is ${GREEN}green${NORMAL}')

    If the terminal doesn't support a given action, then the value of
    the corresponding instance variable will be set to ''.  As a
    result, the above code will still work on terminals that do not
    support color, except that their output will not be colored.
    Also, this means that you can test whether the terminal supports a
    given action by simply testing the truth value of the
    corresponding instance variable:

        >>> term = TerminalController()
        >>> if term.CLEAR_SCREEN:
        ...     print 'This terminal supports clearning the screen.'

    Finally, if the width and height of the terminal are known, then
    they will be stored in the `COLS` and `LINES` attributes.
    """
    # Cursor movement:
    BOL = ''             #: Move the cursor to the beginning of the line
    UP = ''              #: Move the cursor up one line
    DOWN = ''            #: Move the cursor down one line
    LEFT = ''            #: Move the cursor left one char
    RIGHT = ''           #: Move the cursor right one char

    # Deletion:
    CLEAR_SCREEN = ''    #: Clear the screen and move to home position
    CLEAR_EOL = ''       #: Clear to the end of the line.
    CLEAR_BOL = ''       #: Clear to the beginning of the line.
    CLEAR_EOS = ''       #: Clear to the end of the screen

    # Output modes:
    BOLD = ''            #: Turn on bold mode
    BLINK = ''           #: Turn on blink mode
    DIM = ''             #: Turn on half-bright mode
    REVERSE = ''         #: Turn on reverse-video mode
    NORMAL = ''          #: Turn off all modes

    # Cursor display:
    HIDE_CURSOR = ''     #: Make the cursor invisible
    SHOW_CURSOR = ''     #: Make the cursor visible

    # Terminal size:
    COLS = None          #: Width of the terminal (None for unknown)
    LINES = None         #: Height of the terminal (None for unknown)

    # Foreground colors:
    BLACK = BLUE = GREEN = CYAN = RED = MAGENTA = YELLOW = WHITE = ''

    # Background colors:
    BG_BLACK = BG_BLUE = BG_GREEN = BG_CYAN = ''
    BG_RED = BG_MAGENTA = BG_YELLOW = BG_WHITE = ''

    _STRING_CAPABILITIES = """
    BOL=cr UP=cuu1 DOWN=cud1 LEFT=cub1 RIGHT=cuf1
    CLEAR_SCREEN=clear CLEAR_EOL=el CLEAR_BOL=el1 CLEAR_EOS=ed BOLD=bold
    BLINK=blink DIM=dim REVERSE=rev UNDERLINE=smul NORMAL=sgr0
    HIDE_CURSOR=cinvis SHOW_CURSOR=cnorm""".split()
    _COLORS = """BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE""".split()
    _ANSICOLORS = "BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE".split()

    def __init__(self, term_stream=sys.stdout):
        """
        Create a `TerminalController` and initialize its attributes
        with appropriate values for the current terminal.
        `term_stream` is the stream that will be used for terminal
        output; if this stream is not a tty, then the terminal is
        assumed to be a dumb terminal (i.e., have no capabilities).
        """
        # Curses isn't available on all platforms
        try:
            import curses
        except:
            return

        # If the stream isn't a tty, then assume it has no capabilities.
        if not term_stream.isatty():
            return

        # Check the terminal type.  If we fail, then assume that the
        # terminal has no capabilities.
        try:
            curses.setupterm()
        except:
            return

        # Look up numeric capabilities.
        self.COLS = curses.tigetnum('cols')
        self.LINES = curses.tigetnum('lines')

        # Look up string capabilities.
        for capability in self._STRING_CAPABILITIES:
            (attrib, cap_name) = capability.split('=')
            setattr(self, attrib, self._tigetstr(cap_name) or '')

        # Colors
        set_fg = self._tigetstr('setf')
        if set_fg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, color, curses.tparm(set_fg, i) or '')
        set_fg_ansi = self._tigetstr('setaf')
        if set_fg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))), self._ANSICOLORS):
                setattr(self, color, curses.tparm(set_fg_ansi, i) or '')
        set_bg = self._tigetstr('setb')
        if set_bg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, 'BG_' + color, curses.tparm(set_bg, i) or '')
        set_bg_ansi = self._tigetstr('setab')
        if set_bg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))), self._ANSICOLORS):
                setattr(
                    self, 'BG_' + color, curses.tparm(set_bg_ansi, i) or '')

    def _tigetstr(self, cap_name):
        # String capabilities can include "delays" of the form "$<2>".
        # For any modern terminal, we should be able to just ignore
        # these, so strip them out.
        import curses
        cap = curses.tigetstr(cap_name) or b''
        return re.sub(b'\$<\d+>[/*]?', b'', cap)

    def render(self, template):
        """
        Replace each $-substitutions in the given template string with
        the corresponding terminal control string (if it's defined) or
        '' (if it's not).
        """
        return re.sub(b'\$\$|\${\w+}', self._render_sub, template)

    def _render_sub(self, match):
        s = match.group()
        if s == '$$':
            return s
        else:
            return getattr(self, s[2:-1])


def heading(s):
    if term.COLS is not None:
        s = " " + s + " "
        pad = term.COLS - len(s)
        padl = "*" * int(math.floor(pad / 2))
        padr = "*" * int(math.ceil(pad / 2))

        print(term.YELLOW + padl + term.WHITE + s + term.YELLOW + padr + term.NORMAL)
    else:
        print("*" * 20 + s + ("*" * 20))

term = TerminalController()


class RedirectStdStreams(object):

    def __init__(self):
        self.enabled = True

    def __enter__(self):
        if self.enabled:
            self.strio = StringIO()
            self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
            self.old_stdout.flush()
            self.old_stderr.flush()
            sys.stdout, sys.stderr = self.strio, self.strio

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enabled:
            self.strio.flush()
            if exc_type is not None:
                print(self.strio.getvalue())

            sys.stdout = self.old_stdout
            sys.stderr = self.old_stderr
            self.strio.close()
            del self.strio

redirect = RedirectStdStreams()



from __future__ import absolute_import
import contextlib
import random
import string
import sys

import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.query
from sqlalchemy import and_, Table, Column, Integer, Float, ForeignKey
from sqlalchemy.orm import relationship

from .. import core
from .. import temporary_halolist
from .one_hop import HopStrategy

from ..config import num_multihops_max_default as NHOPS_MAX_DEFAULT
from six.moves import range

# Fractional increase or decrease in time to represent future or past when making hops
# (prevents numerical accuracy issues making the db misunderstand a contemporaneous step
# as being in the future or past)
SMALL_FRACTION = 1e-4

class MultiHopStrategy(HopStrategy):
    """An extension of the HopStrategy class that takes multiple hops across
    HaloLinks, up to a specified maximum, before finding the target halo."""

    halo_old = sqlalchemy.orm.aliased(core.halo.Halo, name="halo_old")
    halo_new = sqlalchemy.orm.aliased(core.halo.Halo, name="halo_new")
    timestep_old = sqlalchemy.orm.aliased(core.timestep.TimeStep, name="timestep_old")
    timestep_new = sqlalchemy.orm.aliased(core.timestep.TimeStep, name="timestep_new")

    def __init__(self, halo_from, nhops_max=NHOPS_MAX_DEFAULT, directed=None, target=None,
                 order_by=None, combine_routes=True, min_aggregated_weight=0.0,
                 min_onehop_weight=0.0, min_onehop_reverse_weight=None,
                 include_startpoint=False):
        """Construct a strategy for finding Halos via multiple "hops" along HaloLinks

        :param halo_from:   The halo to start hopping from
        :type halo_from:    core.Halo

        :param nhops_max:   The maximum number of hops to take

        :param directed:    The direction in which to step, which can be
              'backwards' - only take hop if it's backwards in time
              'forwards'  - only take hop if it's forwards in time
              'across'    - only take hop if it's to the same time
              None        - take all available hops

              Note that specifying a direction causes a slight slow-down in each hop because the hop results have
              to be joined to the TimeStep object to work out the direction. But the overall query times are
              often much shorter because the pruning cuts down the number of paths to be explored.

        :param target:      Only return results in target, which can be a TimeStep or Simulation object.
            If None, return all results. Intermediate hops are not restricted by target, only the results returned.

        :param order_by:    Return results in specified order; see HopStrategy for more information

        :param combine_routes:  If True (default), discard multiple routes to the same halo and keep only the
            strongest route. When False, one result is returned for each possible route,
            even if it ultimately reaches the same halo. This can result in a significant
            slow-down.

        :param min_aggregated_weight: Threshold cumulative weight below which routes will be discarded.
              The cumulative weight is defined as the product of the individual hop weights.
              Small values indicate that a small fraction of mass is in common.

        :param min_onehop_weight:     Threshold individual weight below which a route will be discarded.

        :param min_onehop_reverse_weight: Threshold individual weight for the link pointing in the opposite
                                          direction to that being followed, or None for no restriction.
                                          Note that this requires an extra join (if not None) and if there
                                          is no reverse link at all, the result will be dropped.

        :param include_startpoint:    Return the starting halo in the results (default False)
        """
        super(MultiHopStrategy, self).__init__(halo_from, target, order_by)
        self.nhops_max = nhops_max
        self.directed = directed
        self._min_aggregated_weight = min_aggregated_weight
        self._min_onehop_weight = min_onehop_weight
        self._min_onehop_reverse_weight = min_onehop_reverse_weight
        self._include_startpoint = include_startpoint
        self._connection = self.session.connection()
        self._combine_routes = combine_routes

    def temp_table(self):
        """Execute the strategy and return results as a temp_table (see temporary_halolist module)"""
        tt = self._manage_temp_table()
        tt.__enter__()
        try:
            self._generate_multihop_results()
        except:
            tt.__exit__(*sys.exc_info())
            raise
        thl = temporary_halolist.temporary_halolist_table(self.session,
                                                          self._query_ordered.from_self(
                                                              self._link_orm_class.halo_to_id),
                                                          callback=lambda: tt.__exit__(None, None, None)
                                                          )
        return thl

    def _generate_multihop_results(self):
        self._generate_query()
        self._seed_temp_table()
        self._filter_query_for_target(self._target)
        self._make_hops()

    def _execute_query(self):
        with self._manage_temp_table():
            self._generate_multihop_results()
            try:
                results = self._query_ordered.all()
            except sqlalchemy.exc.ResourceClosedError:
                results = []

        self._all = results


    def _supplement_halolink_query_with_filter(self, query, table=None):

        if table is None:
            table = core.halo_data.HaloLink.__table__

        if self._needs_join_for_link_filter():
            query = query. \
                join(self.halo_old, table.c.halo_from_id == self.halo_old.id). \
                join(self.halo_new, table.c.halo_to_id == self.halo_new.id). \
                join(self.timestep_old, self.halo_old.timestep). \
                join(self.timestep_new, self.halo_new.timestep)

        filter = self._generate_link_filter(self.timestep_old, self.timestep_new, table)
        query = query.filter(filter)

        return query

    def _needs_join_for_link_filter(self):
        return self.directed is not None

    def _generate_link_filter(self, timestep_old, timestep_new, table):

        recursion_filter = table.c.weight > self._min_aggregated_weight

        if self.directed is not None:
            directed = self.directed.lower()
            if directed == 'backwards':
                recursion_filter &= timestep_new.time_gyr < timestep_old.time_gyr*(1.0-SMALL_FRACTION)
            elif directed == 'forwards':
                recursion_filter &= timestep_new.time_gyr > timestep_old.time_gyr*(1.0+SMALL_FRACTION)
            elif directed == 'across':
                existing_timestep_ids = self.session.query(core.Halo.timestep_id).\
                    select_from(self._link_orm_class).join(self._link_orm_class.halo_to).distinct()
                recursion_filter &= ~timestep_new.id.in_(existing_timestep_ids)
                recursion_filter &= sqlalchemy.func.abs(timestep_new.time_gyr - timestep_old.time_gyr) < SMALL_FRACTION
            else:
                raise ValueError("Unknown direction %r" % directed)

        return recursion_filter

    def _delete_temp_table(self):
        self._table.drop(checkfirst=True, bind=self._connection)
        self._prelim_table.drop(checkfirst=True, bind=self._connection)
        # self._index.drop(bind=self.connection)
        core.Base.metadata.remove(self._table)
        core.Base.metadata.remove(self._prelim_table)

    def _create_temp_table(self):
        rstr = ''.join(random.choice(string.ascii_lowercase) for _ in range(4))

        multi_hop_link_table = Table(
            'multihoplink_final_' + rstr,
            core.Base.metadata,
            Column('id', Integer, primary_key=True),
            Column('source_id', Integer),
            Column('halo_from_id', Integer, ForeignKey('halos.id')),
            Column('halo_to_id', Integer, ForeignKey('halos.id')),
            Column('weight', Float),
            Column('nhops', Integer),
            prefixes=['TEMPORARY']
        )

        multi_hop_link_prelim_table = Table(
            'multihoplink_prelim_' + rstr,
            core.Base.metadata,
            Column('id', Integer, primary_key=True),
            Column('source_id', Integer),
            Column('halo_from_id', Integer, ForeignKey('halos.id')),
            Column('halo_to_id', Integer, ForeignKey('halos.id')),
            Column('weight', Float),
            Column('nhops', Integer),
            prefixes=['TEMPORARY']
        )

        self._table = multi_hop_link_table
        self._prelim_table = multi_hop_link_prelim_table
        self._table.create(checkfirst=True, bind=self._connection)

        self._prelim_table.create(checkfirst=True, bind=self._connection)

    @contextlib.contextmanager
    def _manage_temp_table(self):
        self._create_temp_table()
        yield
        self._delete_temp_table()

    def _seed_temp_table(self):
        insert_statement = self._table.insert().values(halo_from_id=self.halo_from.id, halo_to_id=self.halo_from.id,
                                    weight=1.0, nhops=0)

        self._connection.execute(insert_statement)

    def _make_hops(self):
        for i in range(0, self.nhops_max):
            generated_count = self._generate_next_level_prelim_links(i)
            if generated_count != 0:
                filtered_count = self._filter_prelim_links_into_final()
            else:
                filtered_count = 0

            if self._hopping_finished(filtered_count):
                break
        self._nhops_taken = i

    def _hopping_finished(self, filtered_count):
        return filtered_count==0

    def _generate_next_level_prelim_links(self, from_nhops=0):

        new_weight = self._table.c.weight * core.halo_data.HaloLink.weight

        if self._combine_routes:
            new_weight = sqlalchemy.func.max(new_weight)

        recursion_query = \
            self.session.query(
                core.halo_data.HaloLink.halo_from_id,
                core.halo_data.HaloLink.halo_to_id.label("halo_to_id"),
                new_weight,
                (self._table.c.nhops + sqlalchemy.literal(1)).label("nhops"),
                self._table.c.source_id). \
                select_from(self._table). \
                join(core.halo_data.HaloLink, and_(self._table.c.nhops == from_nhops,
                                                           self._table.c.halo_to_id == core.halo_data.HaloLink.halo_from_id)). \
                filter(core.halo_data.HaloLink.weight > self._min_onehop_weight
                       )

        if self._combine_routes:
            recursion_query = recursion_query.group_by(core.halo_data.HaloLink.halo_to_id, self._table.c.source_id)

        insert = self._prelim_table.insert().from_select(
            ['halo_from_id', 'halo_to_id', 'weight', 'nhops', 'source_id'],
            recursion_query)

        result = self._connection.execute(insert)

        return result.rowcount

    def _filter_prelim_links_into_final(self):

        q = self.session.query(self._prelim_table.c.halo_from_id,
                               self._prelim_table.c.halo_to_id,
                               self._prelim_table.c.weight,
                               self._prelim_table.c.nhops,
                               self._prelim_table.c.source_id)

        q = self._supplement_halolink_query_with_reverse_hop_filter(q, self._prelim_table)
        q = self._supplement_halolink_query_with_filter(q, self._prelim_table)

        added_rows = self._connection.execute(
            self._table.insert().from_select(['halo_from_id', 'halo_to_id', 'weight', 'nhops', 'source_id'],
                                             q)).rowcount

        self._connection.execute(self._prelim_table.delete())
        return added_rows

    def _supplement_halolink_query_with_reverse_hop_filter(self, query, table=None):
        if self._min_onehop_reverse_weight is None:
            return query

        if table is None:
            table = core.halo_data.HaloLink.__table__

        hl_alias = sqlalchemy.orm.aliased(core.HaloLink)

        query = query. \
                join(hl_alias,
                     and_(hl_alias.halo_from_id==table.c.halo_to_id,
                     hl_alias.halo_to_id==table.c.halo_from_id)).\
                filter(hl_alias.weight>self._min_onehop_reverse_weight)


        return query

    def _construct_orm_class(self):
        rstr = ''.join(random.choice(string.ascii_lowercase) for _ in range(4))
        class_name = "MultiHopHaloLink_"+rstr
        class_base = (core.Base,)
        class_attrs = {"__table__": self._table,
                       "halo_from": relationship(core.halo.Halo, primaryjoin=self._table.c.halo_from_id == core.halo.Halo.id),
                       "halo_to"  : relationship(core.halo.Halo, primaryjoin=(self._table.c.halo_to_id == core.halo.Halo.id)),
                       "source_id" : self._table.c.source_id,
                       "nhops" : self._table.c.nhops
                       }

        return type(class_name,class_base,class_attrs)


    def _generate_query(self):
        self._link_orm_class = self._construct_orm_class()
        self.query = self.session.query(self._link_orm_class)

        if not self._include_startpoint:
            self.query = self.query.filter(self._table.c.nhops>0)

    def _generate_order_arg_from_name(self, name, halo_alias, timestep_alias):
        if name == 'nhops':
            return self._link_orm_class.c.nhops
        else:
            return super(MultiHopStrategy, self)._generate_order_arg_from_name(name, halo_alias, timestep_alias)



from __future__ import absolute_import
import os

import numpy as np

import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data
from tangos import core
from . import translations
from six.moves import range
from six.moves import zip


class HaloStatFile(object):
    """Manages and reads a halo stat file of unspecified format."""
    _id_offset = 0
    _column_translations = {}

    def __new__(cls, timestep):
        for subclass in cls.__subclasses__():
            if subclass.can_load(timestep):
                return object.__new__(subclass)
        else:
            raise IOError("No stat file found for this timestep")

    @classmethod
    def can_load(cls, timestep):
        return os.path.exists(cls.filename(timestep))

    @classmethod
    def filename(cls, timestep):
        raise ValueError("Unknown path to stat file")

    def __init__(self, timestep):
        self._timestep = timestep
        self.filename = self.filename(timestep)

    def iter_rows_raw(self, *args):
        """
        Yield the halo ID and requested columns from each line of the stat file, without any emulation.

        :param args: strings for the column names
        :return: id, arg1, arg2, arg3 where ID is the halo ID and argN is the value of the Nth named column
        """
        with open(self.filename) as f:
            header = self._read_column_names(f)
            ids = [0] + [header.index(a) for a in args]
            for l in f:
                yield self._get_values_for_columns(ids, l)

    def iter_rows(self, *args):
        """
        Yield the halo ID and requested columns from each line of the stat file, emulating the existence of some columns.

        For example, AHF stat files do not contain n_dm; however, its value can be automatically inferred by this
        function. Meanwhile IDL .amiga.stat files rename n_gas as N_gas.

        :param args: strings for the column names
        :return: id, arg1, arg2, arg3 where ID is the halo ID and argN is the value of the Nth named column
        """

        raw_args = []
        for arg in args:
            if arg in self._column_translations:
                raw_args+=self._column_translations[arg].inputs()
            else:
                raw_args.append(arg)

        for raw_values in self.iter_rows_raw(*raw_args):
            values = [raw_values[0]]
            for arg in args:
                if arg in self._column_translations:
                    values.append(self._column_translations[arg](raw_args, raw_values[1:]))
                else:
                    values.append(raw_values[1:][raw_args.index(arg)])
            yield values

    def read(self, *args):
        """Read the halo ID and requested columns from the entire file, returning each column as a separate array"""
        return_values = [[] for _ in range(len(args)+1)]
        for row in self.iter_rows(*args):
            for return_array, value in zip(return_values, row):
                return_array.append(value)

        return [np.array(x) for x in return_values]

    def _get_values_for_columns(self, columns, line):
        results = []
        l_split = line.split()
        for id_this in columns:
            this_str = l_split[id_this]
            if "." in this_str:
                guess_type = float
            else:
                guess_type = int

            try:
                this_cast = guess_type(this_str)
            except ValueError:
                this_cast = this_str

            results.append(this_cast)
        results[0] += self._id_offset
        return results

    def _read_column_names(self, f):
        return [x.split("(")[0] for x in f.readline().split()]

    def make_halo_objects(self, min_NDM=1000):
        halos = []
        import numpy as np
        n_tot = []
        for num, NDM, Nstar, Ngas in self.iter_rows("n_dm", "n_star", "n_gas"):
            n_tot.append(NDM+Nstar+Ngas)
        ids = np.zeros(len(n_tot))
        ids[np.argsort(np.array(n_tot))[::-1]] = np.arange(len(n_tot)) + 1
        cnt = 1
        for num, NDM, Nstar, Ngas in self.iter_rows("n_dm", "n_star", "n_gas"):
            if NDM > min_NDM:
                h = tangos.core.halo.Halo(self._timestep, ids[cnt-1],cnt, NDM, Nstar, Ngas)
                halos.append(h)
            cnt += 1
        return halos

    def add_halos(self, min_NDM=1000):
        halos = self.make_halo_objects(min_NDM)
        session = core.Session.object_session(self._timestep)
        session.add_all(halos)

    def add_halo_properties(self, *property_names):
        halos = self._timestep.halos.all()
        halos_map = {}
        for h in halos:
            halos_map[h.halo_number] = h

        session = core.Session.object_session(self._timestep)

        property_db_names = [tangos.core.dictionary.get_or_create_dictionary_item(session, name) for name in property_names]
        property_objects = []
        for values in self.iter_rows(*property_names):
            halo = halos_map.get(values[0], None)
            if halo is not None:
                for name_object, value in zip(property_db_names, values[1:]):
                    property_objects.append(tangos.core.halo_data.HaloProperty(halo, name_object, value))

        session.add_all(property_objects)


class AHFStatFile(HaloStatFile):
    _id_offset = 1

    _column_translations = {'n_dm': translations.Function(lambda ngas, nstar, npart: npart - ngas - nstar,
                                                          'n_gas', 'n_star', 'npart')}

    @classmethod
    def filename(cls, timestep):
        import glob
        file_list = glob.glob(timestep.filename+'.z*.???.AHF_halos')

        # permit the AHF halos to be in a subfolder called "halos", for yt purposes
        # (where the yt tipsy reader can't cope with the AHF files being in the same folder)
        parts = timestep.filename.split("/")
        parts_with_halo = parts[:-1]+["halos"]+parts[-1:]
        filename_with_halo = "/".join(parts_with_halo)
        file_list+=glob.glob(filename_with_halo+'.z*.???.AHF_halos')

        if len(file_list)==0:
            return "CannotFindAHFHaloFilename"
        else:
            return file_list[0]
        return file


class AmigaIDLStatFile(HaloStatFile):
    _id_offset = 0

    _column_translations = {'n_dm': translations.Rename('N_dark'),
                            'n_gas': translations.Rename('N_gas'),
                            'n_star': translations.Rename("N_star"),
                            'npart': translations.Function(lambda ngas, nstar, ndark: ngas + nstar + ndark,
                                                           "N_dark", "N_gas", "N_star")}

    @classmethod
    def filename(cls, timestep):
        return timestep.filename + '.amiga.stat'




from __future__ import absolute_import
import time
import warnings
import importlib
import sys
import re

import tangos.core.creator
from .. import core, config
import traceback
from six.moves import range

backend = None
_backend_name = config.default_backend
from .. import log
from . import message, jobs, backends


from ..log import logger

def _process_command_line():
    command_line = " ".join(sys.argv)
    match = re.match("^(.*)--backend *=? *([^ ]*)(.*)$", command_line)
    if match is not None:
        global _backend_name
        _backend_name = match.group(2)
        new_command_line = match.group(1)+" "+match.group(3)
        sys.argv = new_command_line.split()

_process_command_line()

def use(name):
    global backend, _backend_name
    if backend is not None:
        warnings.warn("Attempt to specify backend but parallelism is already initialised. This call may have no effect, unless you know exactly what you're doing.", RuntimeWarning)
    _backend_name = name

def init_backend():
    global backend
    if backend is None:
        backend = importlib.import_module("."+_backend_name, package='tangos.parallel_tasks.backends')

def deinit_backend():
    global backend
    backend = None

def parallel_backend_loaded():
    global _backend_name
    return _backend_name!='null'

def launch(function, num_procs=None, args=[]):
    init_backend()
    if _backend_name!='null':
        backend.launch(_exec_function_or_server, num_procs, [function, args])
    else:
        function(*args)

    deinit_backend()

def distributed(file_list, proc=None, of=None):
    """Distribute a list of tasks between all nodes"""

    if type(file_list) == set:
        file_list = list(file_list)

    if _backend_name=='null':
        if proc is None:
            proc = 1
            of = 1
        i = (len(file_list) * (proc - 1)) // of
        j = (len(file_list) * proc) // of - 1
        assert proc <= of and proc > 0
        if proc == of:
            j += 1
        return file_list[i:j + 1]
    else:
        from . import jobs
        return jobs.parallel_iterate(file_list)


def _exec_function_or_server(function, args):
    log.set_identity_string("[%3d] " % backend.rank())
    if backend.rank()==0:
        _server_thread()
    else:
        function(*args)
        MessageExit().send(0)
    _shutdown_parallelism()
    log.set_identity_string("")

class MessageExit(message.Message):
    pass


def _server_thread():
    # Sit idle until request for a job comes in, then assign first
    # available job and move on. Jobs are labelled through the
    # provided iterator

    j = -1
    num_jobs = None
    current_job = None
    alive = [True for i in range(backend.size())]
    awaiting_barrier = [False for i in range(backend.size())]

    while any(alive[1:]):
        obj = message.Message.receive()
        if isinstance(obj, MessageExit):
            alive[obj.source]=False
        else:
            obj.process()


    log.logger.info("Terminating manager")


def _shutdown_parallelism():
    global backend
    log.logger.info("Shutting down parallel_tasks")
    backend.barrier()
    backend.finalize()
    backend = None
    _bankend_name = 'null'




from .lock import RLock
from .barrier import barrier
from . import remote_import


from __future__ import absolute_import
from __future__ import print_function
import glob
import os.path
import fnmatch
from .. import config
from ..log import logger
from six.moves import range
import numpy as np


def find(extension=None, mtd=None, ignore=None, basename="", patterns=[]):
    if mtd == None:
        mtd = config.max_traverse_depth
    if ignore == None:
        ignore = config.file_ignore_pattern
    out = []

    if extension is not None:
        for d in range(mtd + 1):
            out += glob.glob(basename + ("*/" * d) + "*." + extension)

        out = [f[:-(len(extension) + 1)] for f in out]

        out = [f for f in out if not any([fnmatch.fnmatch(f, ipat) for ipat in ignore])]
    else:
        for d in range(mtd + 1):
            for pattern in patterns:
              out += glob.glob(basename + ("*/" * d) + pattern)


    return set(out)


class PatternBasedFileDiscovery(object):
    """Provides methods for pattern-based file discovery, i.e. glob-ing for specific patterns of file in this
    folder and sub-folders"""

    patterns = []  # should be specified by child class. See simulation_output_handlers.pynbody for examples.

    @classmethod
    def best_matching_handler(cls, basename):
        handler_names = []
        handler_timestep_lengths = []
        base = os.path.join(config.base, basename)
        if len(cls.__subclasses__()) == 0:
            return cls
        all_possible_handlers = cls.__subclasses__()
        for possible_handler in all_possible_handlers:
            timesteps_detected = find(basename=base + "/", patterns=possible_handler.patterns)
            handler_names.append(possible_handler)
            handler_timestep_lengths.append(len(timesteps_detected))
        best_handler = handler_names[np.argmax(handler_timestep_lengths)]
        logger.debug("Detected best handler (of %d) is %s",len(all_possible_handlers), best_handler)
        return best_handler

    def enumerate_timestep_extensions(self):
        base = os.path.join(config.base, self.basename)
        extensions = find(basename=base + "/", patterns=self.patterns)
        logger.info("Enumerate timestep extensions base=%r patterns=%s", base, self.patterns)
        for e in extensions:
            if self._is_able_to_load(e):
                yield e[len(base) + 1:]

    def _is_able_to_load(self, fname):
        """Determine whether a named file can be loaded

        Override in child class to filter the pattern-based file matches"""
        return True


from __future__ import absolute_import
import argparse
import contextlib
import gc
import pdb
import random
import sys
import time
import traceback
import warnings
import weakref
import six
import numpy as np
import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm

from .. import properties
from ..util import terminalcontroller, timing_monitor
from .. import parallel_tasks, core
from ..parallel_tasks import database
from ..util.check_deleted import check_deleted
from ..cached_writer import insert_list
from ..log import logger
from six.moves import zip
from tangos import live_calculation



class AttributableDict(dict):
    pass


class PropertyWriter(object):

    def __init__(self):
        self.redirect = terminalcontroller.redirect
        self._writer_timeout = 60
        self._writer_minimum = 60  # don't commit at end of halo if < 1 minute past
        self._current_timestep_id = None
        self._loaded_timestep = None
        self._loaded_halo_id = None
        self._loaded_halo = None

    def _get_parser_obj(self):
        parser = argparse.ArgumentParser()
        core.supplement_argparser(parser)
        parser.add_argument('properties', action='store', nargs='+',
                            help="The names of the halo properties to calculate")
        parser.add_argument('--sims','--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')
        parser.add_argument('--latest', action='store_true',
                            help='Run only on the latest timesteps')
        parser.add_argument('--force', action='store_true',
                            help='Run calculations even if a value is already stored in the database')
        parser.add_argument('--debug', action='store_true',
                            help='Run in debug mode: print rather than creating database entries')
        parser.add_argument('--catch', '--pdb', action='store_true',
                            help='Launch the python debugger when a calculation error occurs')
        parser.add_argument('--backwards', action='store_true',
                            help='Process low-z timesteps first')
        parser.add_argument('--random', action='store_true',
                            help='Process timesteps in random order')
        parser.add_argument('--with-prerequisites', action='store_true',
                            help='Automatically calculate any missing prerequisites for the properties')
        parser.add_argument('--load-mode', action='store', choices=['all', 'partial', 'server', 'server-partial'], required=False, default=None,
                            help="Select a load-mode: " \
                                 "  --load-mode partial:        each node attempts to load only the data it needs; " \
                                 "  --load-mode server:         a server process manages the data;"
                                 "  --load-mode server-partial: a server process figures out the indices to load, which are then passed to the partial loader" \
                                 "  --load-mode all:            each node loads all the data (default, and often fine for zoom simulations).")
        parser.add_argument('--type', action='store', type=str, dest='htype',
                            help="Secify the object type to run on by tag name (or integer). Can be halo, group, or BH.")
        parser.add_argument('--hmin', action='store', type=int, default=0,
                            help="Do not calculate halos below the specified halo")
        parser.add_argument('--hmax', action='store', type=int,
                            help="Do not calculate halos above the specified halo")
        parser.add_argument('--verbose', action='store_true',
                            help="Allow all output from calculations (by default print statements are suppressed)")
        parser.add_argument('--part', action='store', nargs=2, type=int,
                            metavar=('N','M'),
                            help="Emulate MPI by handling slice N out of the total workload of M items. If absent, use real MPI.")
        parser.add_argument('--backend', action='store', type=str, help="Specify the paralellism backend (e.g. pypar, mpi4py)")
        parser.add_argument('--include-only', action='store', type=str, help="Specify a filter that describes which halos the calculation should be executed for.")
        return parser


    def _build_file_list(self):

        query = core.sim_query_from_name_list(self.options.sims)
        files = []
        if self.options.latest:
            for x in query.all():
                try:
                    files.append(x.timesteps[-1])
                except IndexError:
                    pass
        else:
            files = core.get_default_session().query(core.timestep.TimeStep).filter(
                core.timestep.TimeStep.simulation_id.in_([q.id for q in query.all()])). \
                order_by(core.timestep.TimeStep.time_gyr).all()

        if self.options.backwards:
            files = files[::-1]

        if self.options.random:
            random.seed(0)
            random.shuffle(files)

        self.files = files

    def _get_parallel_timestep_iterator(self):
        if self.options.part is not None:
            # In the case of a null backend with manual parallelism, pass the specified part specification
            ma_files = parallel_tasks.distributed(self.files, proc=self.options.part[0], of=self.options.part[1])
        elif self.options.load_mode is not None and self.options.load_mode.startswith('server'):
            # In the case of loading from a centralised server, each node works on the _same_ timestep --
            # parallelism is then implemented at the halo level
            ma_files = self.files
        else:
            # In all other cases, different timesteps are distributed to different nodes
            ma_files = parallel_tasks.distributed(self.files)
        return ma_files

    def _get_parallel_halo_iterator(self, items):
        if self.options.load_mode is not None and self.options.load_mode.startswith('server'):
            # Only in 'server' mode is parallelism undertaken at the halo level. See also
            # _get_parallel_timestep_iterator.

            assert parallel_tasks.backend.size()>1, "Cannot use this load mode outside of a parallel session"

            # First, we need to make a barrier because we can't start writing to the database
            # before all nodes have generated their local work lists
            parallel_tasks.barrier()

            return parallel_tasks.distributed(items)
        else:
            return items

    def parse_command_line(self, argv=None):
        parser = self._get_parser_obj()
        self.options = parser.parse_args(argv)
        core.process_options(self.options)
        self._build_file_list()
        self._compile_inclusion_criterion()

        if self.options.load_mode=='all':
            self.options.load_mode=None

        if self.options.verbose:
            self.redirect.enabled = False

        self.timing_monitor = timing_monitor.TimingMonitor()

    def _compile_inclusion_criterion(self):
        if self.options.include_only:
            self._include = live_calculation.parser.parse_property_name(self.options.include_only)
        else:
            self._include = None


    def _build_halo_list(self, db_timestep):
        query = core.halo.Halo.timestep == db_timestep
        if self.options.htype is not None:
            query = sqlalchemy.and_(query, core.halo.Halo.object_typecode
                                    == core.halo.Halo.object_typecode_from_tag(self.options.htype))

        if self.options.hmin is not None:
            query = sqlalchemy.and_(query, core.halo.Halo.halo_number>=self.options.hmin)

        if self.options.hmax is not None:
            query = sqlalchemy.and_(query, core.halo.Halo.halo_number<=self.options.hmax)

        needed_properties = self._needed_properties()

        halo_query = core.get_default_session().query(core.halo.Halo).order_by(core.halo.Halo.halo_number).filter(query)
        if self._include:
            needed_properties.append(self._include)

        logger.info('Gathering existing properties for all halos in timestep %r',db_timestep)
        halo_query = live_calculation.MultiCalculation(*needed_properties).supplement_halo_query(halo_query)

        halos = halo_query.all()

        if self._include:
            inclusion, = self._include.values(halos)
            if any([not np.issubdtype(inc_i, np.bool) for inc_i in inclusion]):
                raise ValueError("Specified inclusion criterion does not return a boolean")
            if len(inclusion)!=len(halos):
                raise ValueError("Inclusion criterion did not generate results for all the halos")

            # perform filtering:
            halos = [halo_i for halo_i, include_i in zip(halos, inclusion) if include_i]
            logger.info("User-specified inclusion criterion excluded %d of %d halos",
                        len(inclusion)-len(halos),len(inclusion))

        return halos


    def _build_existing_properties(self, db_halo):
        existing_properties = db_halo.all_properties
        need_data = self._needed_property_data()
        existing_properties_data = AttributableDict(
                [(x.name.text, x.data) if x.name.text in need_data
                 else (x.name.text, None) for x in existing_properties])
        existing_properties_data.update(
                [(x.relation.text, x.halo_to) if x.relation.text in need_data
                 else (x.relation.text, None) for x in db_halo.links])
        existing_properties_data.halo_number = db_halo.halo_number
        existing_properties_data.NDM = db_halo.NDM
        existing_properties_data.NGas = db_halo.NGas
        existing_properties_data.NStar = db_halo.NStar
        existing_properties_data['halo_number'] = db_halo.halo_number
        existing_properties_data['finder_id'] = db_halo.finder_id
        return existing_properties_data

    def _build_existing_properties_all_halos(self, halos):
        logger.info("Gathering existing properties...")
        return [self._build_existing_properties(h) for h in halos]

    def _is_commit_needed(self, end_of_timestep, end_of_simulation):
        if len(self._pending_properties)==0:
            return False
        if end_of_simulation:
            return True
        elif end_of_timestep and (time.time() - self._start_time > self._writer_minimum):
            return True
        elif time.time() - self._start_time > self._writer_timeout:
            return True
        else:
            return False


    def _commit_results_if_needed(self, end_of_timestep=False, end_of_simulation=False):

        if self._is_commit_needed(end_of_timestep, end_of_simulation):
            logger.info("Attempting to commit %d halo properties...", len(self._pending_properties))
            insert_list(self._pending_properties)
            logger.info("%d properties were committed", len(self._pending_properties))
            self._pending_properties = []
            self._start_time = time.time()
            self.timing_monitor.summarise_timing(logger)

    def _queue_results_for_later_commit(self, db_halo, names, results, existing_properties_data):
        for n, r in zip(names, results):
            if isinstance(r, properties.ProxyHalo):
                # resolve halo
                r = core.get_default_session().query(core.halo.Halo).filter_by(id=int(r)).first()
            if self.options.force or (n not in list(existing_properties_data.keys())):
                existing_properties_data[n] = r
                if self.options.debug:
                    logger.info("Debug mode - not creating property %r for %r with value %r", n, db_halo, r)
                else:
                    self._pending_properties.append((db_halo, n, r))

    def _needed_properties(self):
        needed = []
        for x in self._property_calculator_instances:
            if isinstance(x.names, six.string_types):
                needed.extend([x.names])
            else:
                needed.extend([name for name in x.names])
            needed.extend([name for name in x.requires_property()])
        return list(np.unique(needed))

    def _needed_property_data(self):
        needed = []
        for x in self._property_calculator_instances:
            needed.extend([name for name in x.requires_property()])
        return list(np.unique(needed))


    def _should_load_halo_particles(self):
        return any([x.requires_particle_data for x in self._property_calculator_instances])

    def _must_load_timestep_particles(self):
        return self._should_load_halo_particles() and self.options.load_mode is None

    def _unload_timestep(self):
        self._loaded_halo = None
        self._current_halo_id = None
        with check_deleted(self._loaded_timestep):
            self._loaded_timestep=None
            self._current_timestep_id = None


    def _set_current_timestep(self, db_timestep):
        if self._current_timestep_id == db_timestep.id:
            return

        self._unload_timestep()

        if self._must_load_timestep_particles():
            self._loaded_timestep = db_timestep.load(mode=self.options.load_mode)

        elif self._should_load_halo_particles():
            # Keep a snapshot alive for this timestep, even if should_load_timestep_particles is False,
            # because it might be needed for the iord's if we are in partial-load mode.
            try:
                self._loaded_timestep = db_timestep.load(mode=self.options.load_mode)
            except IOError:
                pass

        if self.options.load_mode is None:
            self._run_preloop(self._loaded_timestep, db_timestep,
                              self._property_calculator_instances, self._existing_properties_all_halos)
        else:
            self._run_preloop(None, db_timestep,
                              self._property_calculator_instances, self._existing_properties_all_halos)

        self._current_timestep_id = db_timestep.id


    def _set_current_halo(self, db_halo):
        self._set_current_timestep(db_halo.timestep)

        if self._loaded_halo_id==db_halo.id:
            return

        self._loaded_halo_id=db_halo.id
        self._loaded_halo = None

        if self._should_load_halo_particles():
            self._loaded_halo  = db_halo.load(mode=self.options.load_mode)

        if self.options.load_mode is not None:
            self._run_preloop(self._loaded_halo, db_halo.timestep,
                              self._property_calculator_instances, self._existing_properties_all_halos)


    def _get_current_halo_specified_region_particles(self, db_halo, region_spec):
        return db_halo.timestep.load_region(region_spec,self.options.load_mode)

    def _get_halo_snapshot_data_if_appropriate(self, db_halo, db_data, property_calculator):

        self._set_current_halo(db_halo)

        if property_calculator.region_specification(db_data) is not None:
            return self._get_current_halo_specified_region_particles(db_halo, property_calculator.region_specification(db_data))
        else:
            return self._loaded_halo


    def _get_standin_property_value(self, property_calculator):
        if isinstance(property_calculator.names,six.string_types):
            return None
        num = len(property_calculator.names)
        return [None]*num

    def _get_property_value(self, db_halo, property_calculator, existing_properties):
        if property_calculator.no_proxies():
            db_data = db_halo
        else:
            db_data = existing_properties

        result = self._get_standin_property_value(property_calculator)

        try:
            snapshot_data = self._get_halo_snapshot_data_if_appropriate(db_halo, db_data, property_calculator)
        except IOError:
            logger.warn("Failed to load snapshot data for %r; skipping",db_halo)
            self.tracker.register_loading_error()
            return result

        with self.timing_monitor(property_calculator):
            try:
                with self.redirect:
                    result = property_calculator.calculate(snapshot_data, db_data)
                    self.tracker.register_success()
            except Exception:
                self.tracker.register_error()

                if self.tracker.should_log_error(property_calculator):
                    logger.exception("Uncaught exception during property calculation %r applied to %r"%(property_calculator, db_halo))
                    logger.info("Further errors from this calculation on this timestep will be counted but not individually reported.")

                if self.options.catch:
                    tbtype, value, tb = sys.exc_info()
                    pdb.post_mortem(tb)


        return result


    def _run_preloop(self, f, db_timestep, cinstances, existing_properties_all_halos):
        for x in cinstances:
            try:
                with self.redirect:
                    x.preloop(f, db_timestep)
            except Exception:
                logger.exception(
                    "Uncaught exception during property preloop %r applied to %r" % (x, db_timestep))
                if self.options.catch:
                    traceback.print_exc()
                    tbtype, value, tb = sys.exc_info()
                    pdb.post_mortem(tb)


    def run_property_calculation(self, db_halo, property_calculator, existing_properties):
        names = property_calculator.names
        if type(names) is str:
            listize = True
            names = [names]
        else:
            listize = False

        if all([name in list(existing_properties.keys()) for name in names]) and not self.options.force:
            self.tracker.register_already_exists()
            return

        if not property_calculator.accept(existing_properties):
            self.tracker.register_missing_prerequisite()
            return

        results = self._get_property_value(db_halo, property_calculator, existing_properties)

        if listize:
            results = [results]

        self._queue_results_for_later_commit(db_halo, names, results, existing_properties)

    def run_halo_calculation(self, db_halo, existing_properties):
        for calculator in self._property_calculator_instances:
            busy = True
            nloops = 0
            while busy is True:
                busy = False
                try:
                    nloops += 1
                    self.run_property_calculation(db_halo, calculator, existing_properties)
                except sqlalchemy.exc.OperationalError:
                    if nloops > 10:
                        raise
                    time.sleep(1)
                    busy = True

        self._commit_results_if_needed()


    def run_timestep_calculation(self, db_timestep):
        self.tracker = CalculationSuccessTracker()

        logger.info("Processing %r", db_timestep)
        with parallel_tasks.RLock("insert_list"):
            self._property_calculator_instances = properties.instantiate_classes(db_timestep.simulation, self.options.properties)
            if self.options.with_prerequisites:
                self._add_prerequisites_to_calculator_instances(db_timestep)
            db_halos = self._build_halo_list(db_timestep)
            self._existing_properties_all_halos = self._build_existing_properties_all_halos(db_halos)
            core.get_default_session().commit()

        logger.info("Successfully gathered existing properties; calculating halo properties now...")

        logger.info("  %d halos to consider; %d property calculations for each of them",
                    len(db_halos), len(self._property_calculator_instances))

        for db_halo, existing_properties in \
                self._get_parallel_halo_iterator(list(zip(db_halos, self._existing_properties_all_halos))):
            self._existing_properties_this_halo = existing_properties
            self.run_halo_calculation(db_halo, existing_properties)

        logger.info("Done with %r",db_timestep)
        self._unload_timestep()

        self.tracker.report_to_log(logger)
        sys.stderr.flush()

        self._commit_results_if_needed(True)

    def _add_prerequisites_to_calculator_instances(self, db_timestep):
        will_calculate = []
        requirements = []
        for inst in self._property_calculator_instances:
            if isinstance(inst.names, six.string_types):
                will_calculate+=[inst.names]
            else:
                will_calculate+=inst.names
            requirements+=inst.requires_property()

        for r in requirements:
            if r not in will_calculate:
                new_instance = properties.instantiate_class(db_timestep.simulation, r)
                logger.info("Missing prerequisites - added class %r",type(new_instance))
                logger.info("                        providing properties %r",new_instance.names)
                self._property_calculator_instances = [new_instance]+self._property_calculator_instances
                self._add_prerequisites_to_calculator_instances(db_timestep) # everything has changed; start afresh
                break


    def run_calculation_loop(self):
        parallel_tasks.database.synchronize_creator_object()

        self._start_time = time.time()
        self._pending_properties = []

        for f_obj in self._get_parallel_timestep_iterator():
            self.run_timestep_calculation(f_obj)

        self._commit_results_if_needed(True,True)


class CalculationSuccessTracker(object):
    def __init__(self):
        self._skipped_existing = 0
        self._skipped_missing_prerequisite = 0
        self._skipped_error = 0
        self._skipped_loading_error = 0
        self._succeeded = 0

        self._posted_errors = set()

    def should_log_error(self, from_module):
        if from_module not in self._posted_errors:
            self._posted_errors.add(from_module)
            return True
        else:
            return False

    def report_to_log(self, logger):
        logger.info("            Succeeded: %d property calculations", self._succeeded)
        logger.info("              Errored: %d property calculations", self._skipped_error)
        logger.info("  Errored during load: %d property calculations", self._skipped_loading_error)
        logger.info("       Already exists: %d property calculations", self._skipped_existing)
        logger.info("Missing pre-requisite: %d property calculations", self._skipped_missing_prerequisite)

    def register_error(self):
        self._skipped_error+=1

    def register_loading_error(self):
        self._skipped_loading_error+=1

    def register_success(self):
        self._succeeded+=1

    def register_missing_prerequisite(self):
        self._skipped_missing_prerequisite+=1

    def register_already_exists(self):
        self._skipped_existing+=1




from .. import HaloProperties
from ...simulation_output_handlers import yt as yt_handler_module
import numpy as np

class YtHaloProperties(HaloProperties):
    works_with_handler = yt_handler_module.YtOutputSetHandler
    requires_particle_data = True

class HaloDensityProfile(YtHaloProperties):
    """Proof of concept: get halo density profiles from yt"""
    names = "dm_density_profile", "dm_mass_profile"

    def plot_x0(self):
        return self.plot_xdelta()/2

    def plot_xdelta(self):
        return self._simulation.get("approx_resolution_kpc", 0.1)

    def plot_xlabel(self):
        return "r/kpc"

    def plot_ylabel(self):
        return r"$\rho/M_{\odot}\,kpc^{-3}$", r"$M/M_{\odot}$"

    def calculate(self, data, existing_properties):
        # a rough-and-ready profile; faster than the yt built-in profiles which seem to insist on interpolating
        # onto a grid even for radial profiling (is this correct?)

        position = (data["DarkMatter","particle_position"] - data.center).in_units("kpc")
        radius = np.linalg.norm(position,axis=1)
        rmax = radius.max()
        num_bins = int(rmax/self.plot_xdelta())+1

        mass = data["DarkMatter","Mass"].in_units("Msun")

        mass_per_bin,_ = np.histogram(radius, weights=mass, bins=num_bins, range=(0,num_bins*self.plot_xdelta()))

        r = self.plot_x_values(mass_per_bin)
        vol_per_bin = 4*np.pi*r**2*self.plot_xdelta()

        mass_profile = np.cumsum(mass_per_bin)
        den_profile = mass_per_bin/vol_per_bin

        return den_profile, mass_profile




from .spherical_region import SphericalRegionHaloProperties
from .centring import centred_calculation

class BaryonicImages(SphericalRegionHaloProperties):
    names = "gas_map_sideon", "uvi_image_sideon", "gas_map_faceon", "uvi_image_faceon", "gas_map", "uvi_image"

    @classmethod
    def plot_extent(cls):
        return 15.0

    def plot_xlabel(self):
        return "x/kpc"

    def plot_ylabel(self):
        return "y/kpc"

    def plot_clabel(self):
        return r"M$_{\odot}$ kpc$^{-2}$"

    def requires_property(self):
        return ["shrink_center"]

    @centred_calculation
    def calculate(self, particle_data, properties):
        import pynbody.analysis.angmom as angmom
        size = self.plot_extent()
        g, s = self._render_gas(particle_data, size), self._render_stars(particle_data, size)
        with angmom.sideon(particle_data, return_transform=True,
                               cen_size=self._simulation.get("approx_resolution_kpc", 0.1)*10.):
            g_side, s_side = self._render_gas(particle_data, size), self._render_stars(particle_data, size)
            with particle_data.rotate_x(90):
                g_face, s_face = self._render_gas(particle_data, size), self._render_stars(particle_data, size)

        return g_side, s_side, g_face, s_face, g, s

    def _render_projected(self, f, size):
        import pynbody.plot
        im = pynbody.plot.sph.image(f[pynbody.filt.BandPass(
            'z', -size / 2, size / 2)], 'rho', size, units="Msol kpc^-2", noplot=True)
        return im

    def _render_gas(self, f, size):
        if len(f.gas)>0:
            return self._render_projected(f.gas, size)
        else:
            return None

    def _render_stars(self, f, size):
        import pynbody.plot
        if len(f.st)>0:
            return pynbody.plot.stars.render(f.st[pynbody.filt.HighPass('tform',0) & pynbody.filt.BandPass('z', -size / 2, size / 2)],
                                         width=size, plot=False, ret_im=True, mag_range=(16,22))
        else:
            return None



#!/usr/bin/env python2.7

from __future__ import absolute_import
from __future__ import print_function
import tangos as db
import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data
import tangos.core.timestep
import tangos.core.tracking
import tangos.parallel_tasks as parallel_tasks
import tangos.parallel_tasks.database
import tangos.tracking
from tangos.log import logger
import sys
import numpy as np
import gc
import glob
import pynbody
from six.moves import zip


def collect_bh_halos(bh_iord,f, existing_obj_num):
    halo = []
    for bhi in bh_iord:
        bhi = int(bhi)

        eh = np.where(existing_obj_num == bhi)[0]
        if len(eh)==0:
            halo.append(tangos.core.halo.Halo(f, bhi, bhi, 0, 0, 0, 1))

    return halo

def collect_bh_trackers(bh_iord,sim,existing_trackers):
    track = []
    for bhi in bh_iord:
        bhi = int(bhi)
        et = np.where(existing_trackers == bhi)[0]
        if len(et)==0:
            tx = tangos.core.tracking.TrackData(sim, bhi)
            tx.particles = [bhi]
            tx.use_iord = True
            track.append(tx)
    return track


def bh_halo_assign(f_pb):
    f_pbh = f_pb.halos()
    bh_cen_halos=None
    bh_halos = None
    if type(f_pbh) == pynbody.halo.RockstarIntermediateCatalogue:
        bh_cen_halos = f_pbh.get_group_array(family = 'BH')
    if type(f_pbh) == pynbody.halo.AHFCatalogue:
        bh_cen_halos = f_pbh.get_group_array(top_level=False, family = 'bh')
        bh_halos = f_pbh.get_group_array(top_level=True, family='bh')
    if type(f_pbh) != pynbody.halo.AHFCatalogue and type(f_pbh) != pynbody.halo.RockstarIntermediateCatalogue:
        f_pb['gp'] = f_pbh.get_group_array()
        bh_cen_halos = f_pb.star['gp'][np.where(f_pb.star['tform']<0)[0]]
    del f_pbh
    gc.collect()
    return bh_cen_halos, bh_halos


def run():
    session = db.core.get_default_session()
    query = db.sim_query_from_args(sys.argv, session)

    files = db.core.get_default_session().query(tangos.core.timestep.TimeStep).filter(
        tangos.core.timestep.TimeStep.simulation_id.in_([q.id for q in query.all()])). \
        order_by(tangos.core.timestep.TimeStep.time_gyr).all()

    if "backwards" in sys.argv:
        files = files[::-1]

    files = parallel_tasks.distributed(files)
    parallel_tasks.database.synchronize_creator_object(session)

    if "link-only" in sys.argv:
        files = []

    for f in files:
        print(f)
        sim = f.simulation
        
        try:
            f_pb = f.load()
        except:
            print("File not found - continuing")
            continue
        f_pb.physical_units()
        if len(f_pb.star)<1:
            print("No stars - continuing")
            continue

        logger.info("Gathering BH halo information for step %r", f)
        bhobjs, bhobj_nums, bhobj_ids = db.tracker.get_tracker_halos(f)
        halos = f.halos.filter_by(object_typecode=0).all()
        halo_nums = np.array([h.finder_id for h in halos])
        halo_ids = np.array([h.id for h in halos])
        logger.info("Gathering BH - halo link information for step %r", f)
        with parallel_tasks.RLock("bh"):
            bh_dict_id = tangos.core.dictionary.get_or_create_dictionary_item(session, "BH")
            bh_dict_cen_id = tangos.core.dictionary.get_or_create_dictionary_item(session, "BH_central")
            host_dict_id = tangos.core.dictionary.get_or_create_dictionary_item(session, "host_halo")
            session.commit()
        links_c, idf_c, idt_c = db.tracker.get_tracker_links(session, bh_dict_cen_id)
        links, idf, idt = db.tracker.get_tracker_links(session, bh_dict_id)

        logger.info("Gathering BH info from simulation for step %r", f)
        bh_iord = f_pb.star['iord'][np.where(f_pb.star['tform']<0)[0]]
        bh_mass = f_pb.star['mass'][np.where(f_pb.star['tform']<0)[0]]
        bh_iord = bh_iord[np.argsort(bh_mass)[::-1]]
        logger.info("Found %d black holes for %r", len(bh_iord), f)

        logger.info("gathering and committing BHs into step %r", f)
        with session.no_autoflush:
            halo_to_add = collect_bh_halos(bh_iord,f,bhobj_nums)
        with parallel_tasks.RLock("bh"):
            session2 = db.core.Session()
            sim2 = db.get_simulation(sim.id,session2)
            track, track_nums = db.tracker.get_trackers(sim2)
            tracker_to_add = collect_bh_trackers(bh_iord,sim,track_nums)
            session.add_all(tracker_to_add)
            session.add_all(halo_to_add)
            session.commit()
        logger.info("Done committing BH %d trackers and %d halos into %r", len(tracker_to_add), len(halo_to_add), f)

        logger.info("re-gathering bh halo information for %r", f)
        with parallel_tasks.RLock("bh"):
            bhobjs, bhobj_nums, bhobj_ids = db.tracker.get_tracker_halos(f)
        logger.info("Done re-gathering halo information for %r", f)

        logger.info("Getting halo information for BHs from simulation for step %r", f)
        bh_cen_halos, bh_halos = bh_halo_assign(f_pb)
        logger.info("Found associated halos for BHs in %r", f)

        del(f_pb)
        gc.collect()

        logger.info("Gathering BH-Halo links (all) for step %r", f)
        if bh_halos is not None:
            bh_links = []
            bh_halos = bh_halos[np.argsort(bh_mass)[::-1]]
            with session.no_autoflush:
                for bhi, haloi in zip(bh_iord, bh_halos):
                    haloi = int(haloi)
                    bhi = int(bhi)
                    oh = np.where(halo_nums==haloi)[0]
                    obh = np.where(bhobj_nums==bhi)[0]
                    if len(oh)==0:
                        logger.warn("NOTE: skipping BH in halo %d as no corresponding DB object found", haloi)
                        continue
                    if len(obh)==0:
                        logger.warn("WARNING: can't find the db object for BH %d", bhi)
                        continue
                    bhobj_i = bhobjs[obh[0]]
                    h_i = halos[(oh[0])]

                    exists = np.where((idf==halo_ids[oh[0]])&(idt==bhobj_ids[obh[0]]))[0]
                    if len(exists)==0:
                        bh_links.append(tangos.core.halo_data.HaloLink(h_i, bhobj_i, bh_dict_id))

            logger.info("Committing %d HaloLinks for All BHs in each halo for step %r", len(bh_links),f)
            with parallel_tasks.RLock("bh"):
                session.add_all(bh_links)
                session.commit()
            logger.info("Done committing %d links for step %r", len(bh_links), f)

        logger.info("Gathering BH-Halo links (central) for step %r")
        if bh_cen_halos is not None:
            bh_cen_links = []
            bh_cen_halos = bh_cen_halos[np.argsort(bh_mass)[::-1]]
            print("Associated halos: ",bh_cen_halos)
            with session.no_autoflush:
                for bhi, haloi in zip(bh_iord, bh_cen_halos):
                    haloi = int(haloi)
                    bhi = int(bhi)
                    oh = np.where(halo_nums==haloi)[0]
                    obh = np.where(bhobj_nums==bhi)[0]
                    if len(oh)==0:
                        logger.warn("NOTE: skipping BH in halo %d as no corresponding DB object found", haloi)
                        continue
                    if len(obh)==0:
                        logger.warn("WARNING: can't find the db object for BH %d", bhi)
                        continue
                    bhobj_i = bhobjs[obh[0]]
                    h_i = halos[(oh[0])]

                    exists = np.where((idf_c==halo_ids[oh[0]])&(idt_c==bhobj_ids[obh[0]]))[0]
                    if len(exists)==0:
                        bh_cen_links.append(tangos.core.halo_data.HaloLink(h_i, bhobj_i, bh_dict_cen_id))
                        bh_cen_links.append(tangos.core.halo_data.HaloLink(bhobj_i, h_i, host_dict_id))

            logger.info("Committing %d HaloLinks for Central BHs in each halo for step %r", len(bh_cen_links),f)
            with parallel_tasks.RLock("bh"):
                session.add_all(bh_cen_links)
                session.commit()
            logger.info("Done committing %d links for step %r", len(bh_cen_links), f)


def main():
    parallel_tasks.launch(run)




